/* dsr.c
 * identifier: dsr
 * 
 * Dynamic Source Routing 
 * Network layer protocol for the NetSim network simulator. 
 * B. Cameron Lesiuk, 1998
 */ 

/* This module has more than one mode of operation.
 * To have DSR rely ONLY on the network address, leave all of the following
 * options uncommented. This will result in the DSR network layer ONLY 
 * doing network layer routing. No interface caching is done. When a packet
 * needs to be sent, it is simultaneously buffered to be sent on EVERY 
 * interface connected to the node. Furthermore, it's broadcast onto every
 * medium connected to the node, so link-level error detection is disabled.
 * In this mode, the only information used in the in-packet source route
 * (DSRPH_ROUTE) is the pNodeCapID.
 */
/* Option 1: use pSrcInterface and pDstInterface, in conjunction with pNodeCapID. */
#define DSR_OPTION_INTERFACEROUTE
/* Option 2: There is no option 2 ;) */

/* uncomment this for debug messages */
/* #define DSRDEBUG */

#include<stdio.h>
#include<stdlib.h>
#include"memmgr.h"
#include"network.h"
#include"dsr.h"
#include"nam.h"

/**** Settings affecting DSR behaviour etc. ****/
/* You will probably want to adjust these according to the attributes
 * of your medium etc.          
 */
#define DSR_RCENTRY_LIMIT    20   /* size of each route cache entry (restricts your max # hops) */
#define DSR_RCSIZE_LIMIT     2   /* # routes we can hold */
#define DSR_RRCACHE_LIMIT    3   /* max # route request caches */
#define DSR_PACKBUF_LIMIT    1   /* max # packets for packet buffer (ie: ones WE initiated) */
#define DSR_RESPONDBUF_LIMIT 1   /* max # packets for reply buffer (ie: ones we're FORWARDING) */

#define DSR_ROUTE1QUERY_TIMEOUT  NW_MS(200) /* time before query1 is cancelled and query2 started */
#define DSR_ROUTE1QUERY_HOPMAX            1 /* Hop max for query1 */
#define DSR_ROUTE2QUERY_TIMEOUT NW_MS(400) /* time before query2 is cancelled and repeated */
#define DSR_ROUTE2QUERY_HOPMAX           20 /* hop max for query2 */
#define DSR_ROUTE2QUERY_BACKOFF           1 /* backoff multiplier for query 2 repetitions. */
#define DSR_ROUTE2QUERY_TIMEOUTMAX NW_MS(1600) /* max timeout (after backoff) */
#define DSR_FORWARD_HOLDOFFMIN     NW_MS(2) /* holdoff before received packet is forwarded again */
#define DSR_FORWARD_HOLDOFFMAX     NW_MS(4) /* holdoff before received packet is forwarded again */
#define DSR_RRQ_HOLDOFFMIN         NW_MS(2) /* holdoff before a query packet is REPLYed to. */
#define DSR_RRQ_HOLDOFFMAX         NW_MS(4) /* holdoff before a query packet is REPLYed to. */
#define DSR_RRQ_HOPMULT            NW_MS(4) /* added per-hop when REPLYing from cache */

/* Colour - used by NAM */
#define DSR_COLOUR_DATA        2
#define DSR_COLOUR_ROUTEQUERY  0
#define DSR_COLOUR_ROUTEQUERY1 0
#define DSR_COLOUR_ROUTEQUERY2 0
#define DSR_COLOUR_ROUTEREPLY  1
#define DSR_COLOUR_ROUTEERROR  3


/* Statistics constants */
#define DSR_S_PACKETS_SOURCE    0 /* # data packets sent to us for transmission */
#define DSR_S_PACKETS_REPLY     1 /* # data packets WE sent in reply */
#define DSR_S_PACKETS_RECEIVED  2 /* # data packets received (successfully */
#define DSR_S_COUNT 3

/**** Data structures ****/
struct DSRPH_RQ;
struct DSRPH_RE;
struct DSRPH;
struct DSRPH_ROUTE;
struct DSRPACKET;
struct DSRROUTE;
struct DSRSEQNUMCACHE;
struct DSRDATA;

/* Only the DSRPH (DSR Packet Header) structs are actually
 * found in the packets. The other structs are used
 * by the DSR code and held internally at the 
 * node. */

/* HUGE NOTE: Nobody actually has just a DSRPH. In every 
 * NWPACKET.pData region, the packet is formatted as so:
 *  [ DSRPH_* [  DSRPH
 *  [         [  Extra DSRPH_* data
 *  [ DSRPH_ROUTE - variable number of records, as defined by DSRPH.pAddresses above  
 * Consequently, you MUST check DSRPH.pType to know how to typecast
 * the packet's header (as a DSRPH_RQ, _RE, and so on) and also 
 * to find out where the route (DSRPH_ROUTE) records start. Then, 
 * you must check DSRPH.pAddresses to find out how many source route
 * records are present.  */

typedef struct DSRPH {
  long pType;         /* packet type */
  long pAddresses;    /* # addresses in source route */
  long pHopNum;       /* The current hop number (ie: index into path list as to where we should be) */
} DSRPH;

/* packet type-specific data */
typedef struct DSRPH_RQ { /* Route query info */
  struct DSRPH pHeader;
  long pSeqNum; /* assigned by originator */
  long pHopLimit;
  long pDestination; /* desired destination */
} DSRPH_RQ;

typedef struct DSRPH_RR { /* Route reply info */
  struct DSRPH pHeader;
  long pSeqNum; /* assigned by originator */
} DSRPH_RR;

typedef struct DSRPH_RE { /* Route Error info */
  struct DSRPH pHeader;
  long pDestination; /* desired destination */
  long pSender;   /* sender of broken link */
  long pReceiver; /* receiver of broken link */
} DSRPH_RE;

/* This struct is stored in the header of the
 * packet - this is used to store the 
 * source route in the actual packet header.
 * One of these structures will appear 
 * for each hop in the source route. */
typedef struct DSRPH_ROUTE {
  long pNodeCapID;
  long pSrcInterface; /* incoming interface */
  long pDstInterface; /* outgoing interface */
} DSRPH_ROUTE;

typedef struct DSRPACKET {
  long pStatus;
  struct DSRROUTE *pRoute;
  long pDestination;
  NWPACKET *pPacket;
  struct DSRPACKET *pPrev,*pNext;
} DSRPACKET;

typedef struct DSRROUTE {
  long rStatus;     /* status of source route */
  long rTimeout;    /* general purpose stat timeout */
  long rTimeoutVal; /* original timeout value - for geometric backoffs */
  long rDestination;
  long rTimeStamp;  /* time when route was created */
  long rTimeUsed;   /* time when last used */
  DSRPH_ROUTE rHop[DSR_RCENTRY_LIMIT+2];
} DSRROUTE;

typedef struct DSRSEQNUMCACHE {
  long sSource;
  long sDestination;
  long sSequenceNumber;
  long sTimeStamp; /* creation time */
} DSRSEQNUMCACHE;

typedef struct DSRDATA {
  long dSequenceNumber; /* route discovery sequence number for this node */
  struct DSRPACKET *dPackBuf;
  struct NWPACKET *dRespondBuffer;
  struct DSRSEQNUMCACHE dRRCache[DSR_RRCACHE_LIMIT];
  struct DSRROUTE dRoute[DSR_RCSIZE_LIMIT];
  /* Stats */
  long dStats[DSR_S_COUNT];
} DSRDATA;

#define DSR_STATUS_INVALID   0   /* for DSRROUTE/DSRPACKET/etc */
#define DSR_STATUS_ROUTEDIS1 1   /* Route Discovery, level 1 pending */
#define DSR_STATUS_ROUTEDIS2 2   /* Route Discovery, level 2 pending */
#define DSR_STATUS_VALID     3   /* Route/packet/whatever is valid */

#define DSR_PTYPE_DATA       0   /* data packet */
#define DSR_PTYPE_ROUTEQUERY 1   /* route discovery/query */
#define DSR_PTYPE_ROUTEREPLY 2   /* route reply */
#define DSR_PTYPE_ROUTEERROR 3   /* Route error */

/* global counter of how many network convergences we've had (ie:
 * when all nodes have their source routes, the network has converged.) */
int g_dsrConvergence=0;


/* This proc checks to see if the subnetwork has converged. 
 * That is to say, every node wanting a source route has
 * discovered one. */
int dsrCheckForConvergence(NWNODE *p_Node) {
  NWNODE *l_Node;
  DSRDATA *l_Data;
  int l_i,l_HasRoute;

  for (l_Node=p_Node->nSubNet->sNode;l_Node;l_Node=l_Node->nNext) {
    /* Check make sure this is a DSR-operated node */
    if ((l_Node->nTickNode==dsrTickNode)&&(l_Node->nNodeCap->nTickNodeCap)) {
      l_Data=(DSRDATA*)(l_Node->nGeneral);
      l_HasRoute=0;
      for (l_i=0;l_i<DSR_RCSIZE_LIMIT;l_i++) {
        if (l_Data->dRoute[l_i].rStatus==DSR_STATUS_VALID) {
          l_HasRoute=1;
        }
        if ( (l_Data->dRoute[l_i].rStatus==DSR_STATUS_ROUTEDIS1) ||
             (l_Data->dRoute[l_i].rStatus==DSR_STATUS_ROUTEDIS2) ) {
          return 0;
        }
      }
      if (!l_HasRoute) {
        return 0;
      }
    } 
  }
  return 1;
}

/* Increments a stats counter - p_Count should be one of the DSR_S_* constants from wire.h */
void dsrStatCount(NWNODE *p_Node,int p_Count) {
  DSRDATA *l_dsr;
  l_dsr=p_Node->nGeneral;
  l_dsr->dStats[p_Count]++;
}

/* Prints out the stats from the specified subnet */
void dsrStatPrint(NWSUBNETWORK *p_SubNet,int p_Flag) {
  NWNODE *l_Node;
  DSRDATA *l_dsr;
  long l_total[DSR_S_COUNT];
  int l_i;

  for (l_i=0;l_i<DSR_S_COUNT;l_i++)
    l_total[l_i]=0L;

  for (l_Node=p_SubNet->sNode;l_Node;l_Node=l_Node->nNext) {
    l_dsr=l_Node->nGeneral;
    for (l_i=0;l_i<DSR_S_COUNT;l_i++) {
      l_total[l_i]+=l_dsr->dStats[l_i];
      if (p_Flag&NWSPFLAG_CLEAR)
        l_dsr->dStats[l_i]=0L;
    }
  }

  if (p_Flag&NWSPFLAG_PRINT) {
    printf("            data packets:   #Packets  #ReplyPk  Received\n");
    printf("% 3li.%04li STATS proto dsr    ",p_SubNet->sNet->nTick/NWTICK,p_SubNet->sNet->nTick%NWTICK);
    for (l_i=0;l_i<DSR_S_COUNT;l_i++)
      printf("%8li  ",l_total[l_i]);
    printf("\n");
  }
}


/*void dsrCheckRoute(NWNODE *p_Node) {*/
#define dsrCheckRoute(x) ;
void dsrCheckRouteNOTUSED(NWNODE *p_Node) {
  DSRDATA *l_dsr;
  DSRROUTE *l_Route;
  long l_i,l_j;

  l_dsr=p_Node->nGeneral;
  l_Route=l_dsr->dRoute;

  for (l_i=0;l_i<DSR_RCSIZE_LIMIT;l_i++) {
    if (l_Route[l_i].rStatus==DSR_STATUS_VALID) {
      /* check to make sure the first address is our node's address */
      if (l_Route[l_i].rHop[0].pNodeCapID!=p_Node->nNodeCap->nID)
        *((int*)(0x00))=1;

      /* check if the specified destination is in this route */
      for (l_j=0;l_j<DSR_RCENTRY_LIMIT;l_j++) {
        if (l_Route[l_i].rHop[l_j].pNodeCapID==l_Route[l_i].rDestination)
          break;
      }
      if (l_j==DSR_RCENTRY_LIMIT) {
        /* destination not found in path */
        *((int*)(0x00))=2;
      }
    } else {
      /* any non-valid tests ? */
    }

    /* check for repeated routes -- including invalid routes */
    for (l_j=l_i+1;l_j<DSR_RCSIZE_LIMIT;l_j++) {
      if((l_Route[l_i].rDestination==l_Route[l_j].rDestination)&&
         (l_Route[l_i].rDestination!=NWADDR_INVALID))
        *((int*)(0x00))=3;
    }
  }
  return;
}

/* returns 1 if route is being used (ie: reference in dPackBuf) */
long dsrIsRouteUsed(DSRROUTE *p_Route,NWNODE *p_Node) {
  DSRPACKET *l_Pk;
  DSRDATA *l_dsr;

  dsrCheckRoute(p_Node);
  l_dsr=p_Node->nGeneral;
  for (l_Pk=l_dsr->dPackBuf;l_Pk;l_Pk=l_Pk->pNext) {
    if (l_Pk->pRoute==p_Route) 
      return 1;
  }
  return 0;
}

/* finds the specified route, and returns NULL if it doesn't exist */
DSRROUTE *dsrFindRoute(DSRPH_ROUTE p_Dest,NWNODE *p_Node) {
  DSRDATA *l_dsr;
  long l_i,l_j;

  if (p_Dest.pNodeCapID==NWADDR_INVALID)
    return NULL;

  dsrCheckRoute(p_Node);

  l_dsr=p_Node->nGeneral;
  for (l_i=0;l_i<DSR_RCSIZE_LIMIT;l_i++) {
    if (l_dsr->dRoute[l_i].rDestination==p_Dest.pNodeCapID)
      return &(l_dsr->dRoute[l_i]);
  }

  /* check for alternate routes already within other routes */
  for (l_i=0;l_i<DSR_RCSIZE_LIMIT;l_i++) {
    if (l_dsr->dRoute[l_i].rDestination!=NWADDR_INVALID) {
      for(l_j=0;
        l_dsr->dRoute[l_i].rHop[l_j].pNodeCapID!=l_dsr->dRoute[l_i].rDestination;
        l_j++) {
        if (l_dsr->dRoute[l_i].rHop[l_j].pNodeCapID==p_Dest.pNodeCapID)
          return &(l_dsr->dRoute[l_i]);
      }
    }
  } 
  return NULL;
}

/* finds the specified route, and returns a PENDING structure if it doesn't 
 * exist */
DSRROUTE *dsrGetRoute(DSRPH_ROUTE p_Dest,NWNODE *p_Node,long p_Priority) {
  DSRDATA *l_dsr;
  DSRROUTE *l_Route,*l_MasterRoute;
  long l_i;
  long l_Op1;

  dsrCheckRoute(p_Node);
  l_dsr=p_Node->nGeneral;
  l_MasterRoute=NULL;
  /* Priority 0 */
  /* try to find an existing route */
  l_Route=dsrFindRoute(p_Dest,p_Node);
  if (l_Route) {
    if (l_Route->rDestination!=p_Dest.pNodeCapID) {
      /* uh oh, we have a route, but it's a sub-route to another node. 
       * we need to create a new route */
      l_MasterRoute=l_Route;
      l_Route=NULL;
    } else {
      return l_Route; /* only case where we don't have to prep the route */
    }
  }
  p_Priority--;

  /* Priority 1 */
  /* let's find a free route slot to kill and take over */
  if (p_Priority>=0) {
    for (l_i=0;l_i<DSR_RCSIZE_LIMIT;l_i++) {
      if (l_dsr->dRoute[l_i].rDestination==NWADDR_INVALID)  {
        l_Route=&(l_dsr->dRoute[l_i]);
        break;
      }
    }
  }
  p_Priority--;

  /* Priority 2 */
  /* failing that, find oldest unused slot */
  if ((!l_Route)&&(p_Priority>=0)) {
    l_Op1=-1;
    for (l_i=0;l_i<DSR_RCSIZE_LIMIT;l_i++) {
      if (l_dsr->dRoute[l_i].rTimeUsed<0)  {
        if (l_Op1<0) {
          l_Op1=l_i; 
        } else {
          if (l_dsr->dRoute[l_i].rTimeStamp>l_dsr->dRoute[l_Op1].rTimeStamp)
            l_Op1=l_i;
        }
      }
    }
    if (l_Op1>=0) {
      l_Route=&(l_dsr->dRoute[l_Op1]);
    }
  }
  p_Priority--;

  /* Priority 3 */
  if ((!l_Route)&&(p_Priority>=0)) {
    /* failing that, find oldest slot not currently being used */
    l_Op1=-1;
    for (l_i=0;l_i<DSR_RCSIZE_LIMIT;l_i++) {
      /* first check if this route is used by any packets in our
       * packet buffer */
      if (!dsrIsRouteUsed(l_Route,p_Node)) { 
        if (l_Op1<0) {
          l_Op1=l_i;
        } else {
          if (l_dsr->dRoute[l_i].rTimeUsed>l_dsr->dRoute[l_Op1].rTimeUsed)
            l_Op1=l_i;
        }
      }
    }
    if (l_Op1>=0) {
      l_Route=&(l_dsr->dRoute[l_Op1]);
    }
  }
  /* Prep the route */
  dsrCheckRoute(p_Node);
  if (l_Route) {
    if (l_MasterRoute&&(l_MasterRoute!=l_Route)) {
      /* we have a template to copy!! */
      l_Route->rTimeUsed=-1;
      l_Route->rTimeStamp=l_MasterRoute->rTimeStamp;
      l_Route->rDestination=p_Dest.pNodeCapID;
      l_Route->rStatus=DSR_STATUS_VALID;
      l_i=-1;
      do { 
        l_i++;
	/* copy entire route structure */
        l_Route->rHop[l_i]=l_MasterRoute->rHop[l_i];
      } while((l_MasterRoute->rHop[l_i].pNodeCapID!=p_Dest.pNodeCapID)&&(l_i<DSR_RCENTRY_LIMIT));
      if (l_i>=DSR_RCENTRY_LIMIT) {
        l_Route->rStatus=DSR_STATUS_INVALID;
       /*  *((int*)(0x00))=8;*/
#ifdef DSR_OPTION_INTERFACEROUTE
      } else {
        /* invalidate last link's out-going interface */
        l_Route->rHop[l_i].pDstInterface=NWADDR_INVALID;
#endif
      }
    } else {
      l_Route->rTimeUsed=-1;
      l_Route->rTimeStamp=p_Node->nNodeCap->nNet->nTick;
      l_Route->rDestination=p_Dest.pNodeCapID;
      l_Route->rStatus=DSR_STATUS_INVALID;
    }
  } 
  dsrCheckRoute(p_Node);
  return l_Route;
}

void dsrAddRouteSFirst(DSRPH_ROUTE p_Dest, DSRPH_ROUTE *p_Route, NWNODE *p_Node) {
  DSRDATA *l_dsr;
  long l_i;
  DSRROUTE *l_Route;

  dsrCheckRoute(p_Node);
  l_dsr=p_Node->nGeneral;
  l_Route=dsrFindRoute(p_Dest,p_Node);
  if (l_Route) {
    /* If we've already got a route, but they don't share destinations, 
     * we just assume this (longer) route should stay so we don't modify it.
     * A bad decision for many reasons, but easy to implement. */
    if (l_Route->rDestination!=p_Dest.pNodeCapID)
      return;
    if (l_Route->rStatus!=DSR_STATUS_VALID) {
      /* trade in right now -- no questions asked */
      l_Route->rStatus=DSR_STATUS_VALID;
      l_Route->rTimeStamp=p_Node->nNodeCap->nNet->nTick;
      l_i=-1;
      do {
        l_i++;
	/* copy over all info: pNodeCapID, pSrcInterface, pDstInterface */
        l_Route->rHop[l_i]=p_Route[l_i];
      } while((p_Route[l_i].pNodeCapID!=p_Dest.pNodeCapID)&&(l_i<DSR_RCENTRY_LIMIT));
      if (l_i>=DSR_RCENTRY_LIMIT) {
        l_Route->rStatus=DSR_STATUS_INVALID;
        /*  *((int*)(0x00))=8; */
      }
      dsrCheckRoute(p_Node);
      return;
    }
    /* compare routes */
    for (l_i=0;
      (p_Route[l_i].pNodeCapID!=p_Dest.pNodeCapID)&&(l_Route->rHop[l_i].pNodeCapID!=p_Dest.pNodeCapID)&&(l_i<DSR_RCENTRY_LIMIT);
      l_i++);
    if (l_Route->rHop[l_i].pNodeCapID==p_Dest.pNodeCapID) 
      return;
    /* swap in new route */
    l_i=-1;
    do {
      l_i++;
      /* copy over all info: pNodeCapID, pSrcInterface, pDstInterface */
      l_Route->rHop[l_i]=p_Route[l_i];
    } while((p_Route[l_i].pNodeCapID!=p_Dest.pNodeCapID)&&(l_i<DSR_RCENTRY_LIMIT));
    if (l_i>=DSR_RCENTRY_LIMIT)
      *((int*)(0x00))=8;
    l_Route->rTimeStamp=p_Node->nNodeCap->nNet->nTick;
  } else {
    /* new route */
    /* should we add it? */
    l_Route=dsrGetRoute(p_Dest,p_Node,2); /* just use priority 2 for now */
    if (l_Route) {
      /* populate new route with info */
      l_i=-1;
      do {
        l_i++;
	/* copy over all info: pNodeCapID, pSrcInterface, pDstInterface */
        l_Route->rHop[l_i]=p_Route[l_i];
      } while ((p_Route[l_i].pNodeCapID!=p_Dest.pNodeCapID)&&(l_i<DSR_RCENTRY_LIMIT));
      if (l_i>=DSR_RCENTRY_LIMIT) {
        l_Route->rStatus=DSR_STATUS_INVALID;
        /* *((int*)(0x00))=8;*/
      } else {
        l_Route->rStatus=DSR_STATUS_VALID;
      }
    }
  }
  dsrCheckRoute(p_Node);
  return;
}

void dsrAddRouteDFirst(DSRPH_ROUTE *p_Route,NWNODE *p_Node) {
  DSRPH_ROUTE l_Route[DSR_RCENTRY_LIMIT];
  long l_i,l_j;

  dsrCheckRoute(p_Node);
  /* first, count our hops */
  for (l_i=0;
    (p_Route[l_i].pNodeCapID!=p_Node->nNodeCap->nID)&&(l_i<DSR_RCENTRY_LIMIT);
    l_i++);
  for (l_j=0;l_j<=l_i;l_j++) {
    l_Route[l_j].pNodeCapID=p_Route[l_i-l_j].pNodeCapID;
#ifdef DSR_OPTION_INTERFACEROUTE
    /* reverse the interface order */
    l_Route[l_j].pSrcInterface=p_Route[l_i-l_j].pDstInterface;
    l_Route[l_j].pDstInterface=p_Route[l_i-l_j].pSrcInterface;
#endif
  }
  dsrAddRouteSFirst(p_Route[0], l_Route, p_Node);
  dsrCheckRoute(p_Node);
  return;
}

void dsrRemoveRoute(long p_Node1,long p_Node2,NWNODE *p_Node) {
  long l_i,l_j;
  DSRROUTE *l_Route;
  DSRDATA *l_dsr;

  dsrCheckRoute(p_Node);
  l_dsr=p_Node->nGeneral;

  /* check all our cached routes for this pair -- and remove them
   * if they exist */
  for (l_i=0;l_i<DSR_RCSIZE_LIMIT;l_i++) {
    /* Check this route */
    l_Route=&(l_dsr->dRoute[l_i]);
    if (l_Route->rDestination!=NWADDR_INVALID) {
      if (l_Route->rStatus==DSR_STATUS_VALID) {
        /* check this sucker for this hop */
	l_j=0;
	do {
	  l_j++; 
	  if (l_j>=DSR_RCENTRY_LIMIT)
	    break;
          if (l_Route->rHop[l_j].pNodeCapID==p_Node1) {
            if (l_Route->rHop[l_j-1].pNodeCapID==p_Node2) {
              l_Route->rStatus=DSR_STATUS_INVALID;
              break;
            }
          } else if (l_Route->rHop[l_j].pNodeCapID==p_Node2) {
            if (l_Route->rHop[l_j-1].pNodeCapID==p_Node1) {
              if (dsrIsRouteUsed(l_Route,p_Node)) {
                /* used -- just mark as invalid to repeat source lookup */
                if (l_Route->rStatus==DSR_STATUS_VALID) {
                  l_Route->rStatus=DSR_STATUS_INVALID;
                } /* else leave it doing it's source lookup already in progress */
              } else {
                l_Route->rDestination=NWADDR_INVALID;
              }
              break;
            }
          }
        } while (l_Route->rHop[l_j].pNodeCapID!=l_Route->rDestination);
      }
    }
  }
  dsrCheckRoute(p_Node);
  return;
}

DSRPH_ROUTE *dsrIndexToRoute(NWPACKET *p_Packet) {
  DSRPH *l_ph; /* packet header structure */
  DSRPH_ROUTE *l_addr;

  if (!p_Packet) return NULL;
  l_ph=((DSRPH*)(p_Packet->pData));
  if (l_ph->pType==DSR_PTYPE_DATA) 
    l_addr=((DSRPH_ROUTE*)(((void*)(l_ph))+sizeof(DSRPH)));
  else if (l_ph->pType==DSR_PTYPE_ROUTEQUERY) 
    l_addr=((DSRPH_ROUTE*)(((void*)(l_ph))+sizeof(DSRPH_RQ)));
  else if (l_ph->pType==DSR_PTYPE_ROUTEREPLY) 
    l_addr=((DSRPH_ROUTE*)(((void*)(l_ph))+sizeof(DSRPH_RR)));
  else if (l_ph->pType==DSR_PTYPE_ROUTEERROR) 
    l_addr=((DSRPH_ROUTE*)(((void*)(l_ph))+sizeof(DSRPH_RE)));
  else {
    printf("OH FUCK!\n");
    *((int*)(0x00))=1;
    return NULL;
  }
  return l_addr;
}

/* This duplicates the following dsrGetNextHop in large part. I hate 
 * to have to do it, but I didn't do the original design very well, 
 * so this duplication of code is necessary. */
/* This proc finds the given node in the source route, and returns
 * the index to it. */
long dsrFindNodeInRoute(NWPACKET *p_Packet,NWNODE *p_Node) {
  DSRPH *l_ph; /* packet header structure */
  DSRPH_ROUTE *l_addr;
  long l_aLimit,l_i;

  l_ph=((DSRPH*)(p_Packet->pData));
  l_aLimit=l_ph->pAddresses;
  l_addr=dsrIndexToRoute(p_Packet);
  for (l_i=0;l_i<l_aLimit;l_i++)
    if (l_addr[l_i].pNodeCapID==p_Node->nNodeCap->nID)
      break;

  if (l_i==l_aLimit)
    return -1;

  return l_i;
}

/* Checks for the next hop in the address (assuming we're the current hop)
 * and verifies we're in the CORRECT spot for the packet as well (so we
 * don't go forwarding packets we've overheard) */
long dsrGetNextHop(NWPACKET *p_Packet,NWNODE *p_Node) {
  DSRPH *l_ph; /* packet header structure */
  DSRPH_ROUTE *l_addr;
  long l_aLimit,l_i;

  l_ph=((DSRPH*)(p_Packet->pData));
  l_aLimit=l_ph->pAddresses;
  l_addr=dsrIndexToRoute(p_Packet);
  for (l_i=0;l_i<l_aLimit;l_i++)
    if (l_addr[l_i].pNodeCapID==p_Node->nNodeCap->nID)
      break;

  if (l_i==l_aLimit)
    return NWADDR_INVALID;

  /* we found US. Check if we're where we're supposed to be. */
  if (l_ph->pHopNum!=l_i)
    return NWADDR_INVALID;  /* we've overheard a packet not meant for us */

  /* now advance to NEXT hop */
  l_i++;
  if (l_i<l_aLimit) {
    return (l_addr[l_i].pNodeCapID);
  }
  /* oops, we're past the end of our buffer! How did that happen ? */
  /* this can be used to check if the specified node is in a source route or 
   * not. */
  return NWADDR_INVALID;
}

void dsrInsertRespondPacket(NWPACKET *p_Packet, NWNODE *p_Node) {
  DSRDATA   *l_dsr;
  NWPACKET *l_Packet;

  /* printf("         DSR:     Respond packet inserted...\n"); */
  l_dsr=p_Node->nGeneral;
  /* now put it at the end of the respond buffer */
  if (l_dsr->dRespondBuffer) {
    for (l_Packet=l_dsr->dRespondBuffer;
      l_Packet->pNext;
      l_Packet=l_Packet->pNext);
    l_Packet->pNext=p_Packet;
    p_Packet->pPrev=l_Packet;
  } else {
    l_dsr->dRespondBuffer=p_Packet;
  } 
}

/* Counts and returns the size of the packet buffer for the given node.
 * ie: counts how many packets we have in our buffer which WE originated. */
long dsrPacketBufSize(NWNODE *p_Node) {
  DSRDATA *l_dsr;
  DSRPACKET *l_Pk;
  long l_i;

  l_dsr=p_Node->nGeneral;
  /* count the number of packets already in our queue to check for overflow */
  for (l_i=0,l_Pk=l_dsr->dPackBuf;
    l_Pk;
    l_i++,l_Pk=l_Pk->pNext);
  return l_i;
}

/* Counts and returns the size of the respond packet buffer for the given node.
 * ie: counts how many packets we have in our buffer which we are forwarding. */
long dsrRespondBufSize(NWNODE *p_Node) {
  DSRDATA *l_dsr;
  NWPACKET *l_Pk;
  long l_i;

  l_dsr=p_Node->nGeneral;
  /* count the number of packets already in our queue to check for overflow */
  for (l_i=0,l_Pk=l_dsr->dRespondBuffer;
    l_Pk;
    l_i++,l_Pk=l_Pk->pNext);
  return l_i;
}

NWPACKET *dsrAllocRespondPacket(NWPACKET *p_Packet, NWNODE *p_Node) {
  DSRDATA   *l_dsr;
  NWPACKET *l_Packet;

  l_dsr=p_Node->nGeneral;
  l_Packet=nwCopyPacket(p_Packet,1);

  /* forward on to next hop - insert packet into respond buffer */
  /* set up the holdoff value */
  l_Packet->pHoldoff=
    rand()%(DSR_FORWARD_HOLDOFFMAX-DSR_FORWARD_HOLDOFFMIN)
    +DSR_FORWARD_HOLDOFFMIN;
  /* now put it at the end of the respond buffer */
  dsrInsertRespondPacket(l_Packet, p_Node);
  return l_Packet;
}

DSRPACKET *dsrAllocPacket(NWPACKET *p_Packet,NWNODE *p_Node) {
  DSRPACKET *l_Pk,*l_Pk2;
  DSRDATA   *l_dsr;

  l_Pk=mmuAlloc(sizeof(DSRPACKET));
  if (l_Pk) {
    l_dsr=p_Node->nGeneral;

    /* init packet structure */
    l_Pk->pStatus=DSR_STATUS_INVALID;
    l_Pk->pRoute=NULL;
    l_Pk->pDestination=p_Packet->pDestID;
    l_Pk->pPacket=nwCopyPacket(p_Packet,0);

    /* insert this packet into the node's linked list */
    if (l_dsr->dPackBuf) {
      /* put it at the end */
      l_Pk2=l_dsr->dPackBuf;
      while(l_Pk2) {
        l_Pk->pPrev=l_Pk2;
        l_Pk2=l_Pk2->pNext;
      }
      l_Pk->pPrev->pNext=l_Pk;
      l_Pk->pNext=NULL;
    } else {
      l_Pk->pPrev=l_Pk->pNext=NULL;
      l_dsr->dPackBuf=l_Pk;
    }
  }
  return l_Pk;
}

void dsrFreePacket(DSRPACKET *p_dsrPacket,NWNODE *p_Node) {
  DSRDATA   *l_dsr;

  l_dsr=p_Node->nGeneral;

  /* first, free attached NWPACKET */
  nwFreePacket(p_dsrPacket->pPacket);

  /* next, unlink from the buffer */
  if (p_dsrPacket->pPrev) {
    p_dsrPacket->pPrev->pNext=p_dsrPacket->pNext;
  } else {
    l_dsr->dPackBuf=p_dsrPacket->pNext;
  }
  if (p_dsrPacket->pNext) {
    p_dsrPacket->pNext->pPrev=p_dsrPacket->pPrev;
  }

  /* finally, free this structure */
  mmuFree(p_dsrPacket);

  /* note we dont' worry about the route -- it's just a link
   * and the routes are managed elsewhere */
}

void dsrInitNode(NWNODE *p_Node) {
  DSRDATA *l_dsr;
  long l_i;

  l_dsr=mmuAlloc(sizeof(DSRDATA));
  p_Node->nGeneral=l_dsr;

  l_dsr->dPackBuf=NULL;
  l_dsr->dRespondBuffer=NULL;
  l_dsr->dSequenceNumber=0L;
  for (l_i=0;l_i<DSR_RRCACHE_LIMIT;l_i++) {
    l_dsr->dRRCache[l_i].sSource=NWADDR_INVALID;
    l_dsr->dRRCache[l_i].sDestination=NWADDR_INVALID;
  } 

  for (l_i=0;l_i<DSR_RCSIZE_LIMIT;l_i++) {
    l_dsr->dRoute[l_i].rStatus=DSR_STATUS_INVALID;
    l_dsr->dRoute[l_i].rDestination=NWADDR_INVALID;
  } 

  for (l_i=0;l_i<DSR_S_COUNT;l_i++) 
    l_dsr->dStats[l_i]=0L;

  return;
}

void dsrTickNode(NWNODE *p_Node) {
  NWPACKET *l_Packet,*l_OrigPk;
  DSRPACKET *l_dsrPacket;
  NWINTERFACE *l_Int;
  DSRDATA *l_dsr;
  DSRROUTE *l_Route;
  DSRPH_ROUTE *l_Dest=NULL,l_Tmp;
  DSRPH *l_dsrPH;
  DSRPH_RQ *l_dsrPH_RQ;
  long l_i;
  long l_BroadcastOK;
/*  long l_k;*/
  long l_SrcIntID=0,l_DstIntID=0;
  long l_SentPacket;

  l_dsr=p_Node->nGeneral;

  mmuSanityCheck();

  /* do any housekeeping for routes, timeouts, etc here */

  /* update holdoffs in our respond queue */
  for (l_Packet=l_dsr->dRespondBuffer;l_Packet; l_Packet=l_Packet->pNext) {
    if (l_Packet->pHoldoff)
      l_Packet->pHoldoff--;
  }
  /* update the timeouts on our route cache */
  for (l_i=0;l_i<DSR_RCSIZE_LIMIT;l_i++) {
    if ((l_dsr->dRoute[l_i].rStatus!=DSR_STATUS_INVALID)&&
        (l_dsr->dRoute[l_i].rDestination!=NWADDR_INVALID)&&
        (l_dsr->dRoute[l_i].rTimeout))
      l_dsr->dRoute[l_i].rTimeout--;
  }

  /* check if we have a new packet to send */
  for (l_OrigPk=p_Node->nNodeCap->nPacketBuffer;
    l_OrigPk;
    l_OrigPk=l_OrigPk->pNext) {
    if ((l_OrigPk->pPayloadLen==32)||(l_OrigPk->pPayloadLen==1000)) 
      dsrStatCount(p_Node,DSR_S_PACKETS_SOURCE);
    /* add to queue on this node - since we've only got one buffer, always add to first medium's queue. */
    /* However, we have to fill out some extra info before we do this */
    l_OrigPk->pNamConversationID=l_OrigPk->pSourceID;
    l_OrigPk->pNamColour=DSR_COLOUR_DATA;
    strcpy(l_OrigPk->pNamString,"DSR-D");
    namPacketQueueAdd(p_Node,l_OrigPk);
    if (dsrPacketBufSize(p_Node)<DSR_PACKBUF_LIMIT) {
      /* there's room in our pPackBuf for this sucker - let's add it in */
      l_dsrPacket=dsrAllocPacket(l_OrigPk,p_Node);
      l_Packet=l_dsrPacket->pPacket;

      /* OK, let's set up the state for this packet */
      /* first, lets search our route database -- see if we
       * have an existing route available */
      l_Tmp.pNodeCapID=l_dsrPacket->pDestination;
      l_Route=dsrGetRoute(l_Tmp,p_Node,99);
      if (l_Route) {
#ifdef DSRDEBUG
        printf("%8li DSR: Packet %li from %li to %li buffered.\n",
          p_Node->nNodeCap->nNet->nTick,
  	  l_Packet->pID,
          l_Packet->pSourceID,l_Packet->pDestID);
#endif
        /* Great, we can tag this packet as having a route. We'll inherit
         * the route's status */
        l_dsrPacket->pStatus=DSR_STATUS_VALID;
        l_dsrPacket->pRoute=l_Route;
        l_Route->rTimeUsed=p_Node->nNodeCap->nNet->nTick;
      } else {
        /* uh oh. Let's just mark this packet accordingly. */
        l_dsrPacket->pStatus=DSR_STATUS_INVALID;
        l_dsrPacket->pRoute=NULL;
        *((int*)(0x00))=2;
      }
      /* that's it. The packet is in our queue. It'll get processed
       * when it's appropriate. */
      /* DONE! */
    } else {
      /* there's no room for this packet in our pPackBuf!! OVERFLOW!! We lost this packet */
      namPacketQueueDiscard(p_Node,l_OrigPk);
      /* this is a bit odd, but we don't need to free l_OrigPk because that's done
       * elsewhere by the traffic generator or something in the NodeCap logic */
    }
  }

  /* check if our interface is free to send a new packet. */
  /* Note that EVERY interface must be free and ready to transmit 
   * in order to transmit a broadcast (NWADDR_BROADCAST) packet
   * or else we won't transmit on ANY interface. We do this to 
   * avoid per-interface buffering or some such thing. An efficient
   * way of earmarking what's been sent to what interface could be
   * done, but I'm shooting for a fairly stupid DSR implementation here. */
  /* note we transmit packets in the order they were 
   * buffered (unless they have a holdoff). One exception 
   * is if we hit a packet that needs to be broadcast, we'll 
   * not search for any more packets. IE: We'll stop transmitting
   * until all interfaces are clear and the broadcast has a chance
   * to be sent. */
  mmuSanityCheck();

  /* The following just checks if all interfaces are clear (in case we hit a broadcast packet) */
  l_BroadcastOK=1;
  for (l_Int=p_Node->nInterface;l_Int;l_Int=l_Int->iNodeNext) {
    if (!((l_Int->iMedium->mSendPacket)(l_Int,NULL,NWADDR_INVALID))) 
      l_BroadcastOK=0;
  }

  /* respond buffer gets first dibbs */
  /* Repond packets are either protocol responses or packet forwarding. */
  /* Send the packet at the top of the list, or the first one with 
   * a holdoff of 0.  */
  if (l_dsr->dRespondBuffer) {
#ifdef DSR_OPTION_INTERFACEROUTE
    for (l_Packet=l_dsr->dRespondBuffer;
      (l_Packet);
      l_Packet=l_Packet->pNext) {
      /* Check if this packet can be transmitted. If it can, exit loop. */
      if (!(l_Packet->pHoldoff)) {
        /* let's extract our source and destination interface for this packet */
	l_Dest=dsrIndexToRoute(l_Packet);
        l_Dest=&(l_Dest[dsrFindNodeInRoute(l_Packet,p_Node)]);
	/* l_Dest now points to OUR record in the source route */
        l_SrcIntID=l_Dest->pDstInterface;
        if (l_SrcIntID==NWADDR_BROADCAST) {
	  /* we should only get this situation for a route QUERY packet */
	  l_DstIntID=NWADDR_BROADCAST;
	  if (l_BroadcastOK)
	    break; /* we're ready to broadcast this packet! */
	} else {
	  l_Dest++; /* advance to next record - the NEXT HOP!! */
	  l_DstIntID=l_Dest->pSrcInterface;
          /* ok, we've got our interfaces, now check if the appropriate 
	   * interface is FREE for transmission. First we have to FIND the interface. */
          for (l_Int=p_Node->nInterface;l_Int;l_Int=l_Int->iNodeNext) {
            if (l_Int->iID==l_SrcIntID)
  	    break;
          }
	  if (l_Int) {
	    /* let's check if this interface is free for transmission */
	    if ((l_Int->iMedium->mSendPacket)(l_Int,NULL,NWADDR_INVALID))
	      break; /* It's FREE! Let's get out of this loop! */
	  } else {
	    /* a nasty error has occurred... ouch!! */
            printf("OH FUCK!\n");
            *((int*)(0x00))=8;
	  }
	}
      } /* else this packet has a holdoff - ignore for now */
    }

#else /* DSR_OPTION_INTERFACEROUTE */
    /* Default option is to broadcast every packet on every medium */
    l_SrcIntID=NWADDR_BROADCAST;
    l_DstIntID=NWADDR_BROADCAST; 
    if (l_BroadcastOK) {
      for (l_Packet=l_dsr->dRespondBuffer;
           (l_Packet)&&(l_Packet->pHoldoff);
           l_Packet=l_Packet->pNext) ;
     } else {
       l_Packet=NULL;
     } 
#endif /* DSR_OPTION_INTERFACEROUTE */

    if (l_Packet) {
      /* let's unlink this packet */
      if (l_Packet->pPrev) {
        l_Packet->pPrev->pNext=l_Packet->pNext;
      } else {
        l_dsr->dRespondBuffer=l_Packet->pNext;
      }
      if (l_Packet->pNext)
        l_Packet->pNext->pPrev=l_Packet->pPrev;
      /* and cut this packet off from the list as well. */
      l_Packet->pNext=l_Packet->pPrev=NULL;
      /* remove the packet from the queue (NAM) */
      namPacketQueueRemove(p_Node,l_Packet);
      /* now send the packet */
      for (l_Int=p_Node->nInterface;l_Int;l_Int=l_Int->iNodeNext) {
        if ((l_Int->iID==l_SrcIntID)||(l_SrcIntID==NWADDR_BROADCAST)) {
#ifdef DSRDEBUG
          printf("%8li DSR: Sending packet %li from node %li (%li to %li) via Medium %li (type %s)\n",
            p_Node->nNodeCap->nNet->nTick,
            l_Packet->pID,
    	    l_Int->iNode->nNodeCap->nID,
            l_Packet->pSourceID,l_Packet->pDestID,
	    l_Int->iMedium->mID,
            (((DSRPH*)(l_Packet->pData))->pType==DSR_PTYPE_DATA)?"DATA":
            (((DSRPH*)(l_Packet->pData))->pType==DSR_PTYPE_ROUTEQUERY)?"QUERY":
            (((DSRPH*)(l_Packet->pData))->pType==DSR_PTYPE_ROUTEREPLY)?"REPLY":
            (((DSRPH*)(l_Packet->pData))->pType==DSR_PTYPE_ROUTEERROR)?"ERROR":
            "**???**");
#endif
#ifdef DSR_OPTION_INTERFACEROUTE
          if (l_SrcIntID==NWADDR_BROADCAST) {
	    /* we have to save our outgoing interface ID in the route
	     * record. Fortunately, l_Dest should still be valid. */
            l_Dest->pDstInterface=l_Int->iID;
	  }
#endif
          (l_Int->iMedium->mSendPacket)(l_Int,l_Packet,l_DstIntID);
	}
      }
      /* lastly, let's free the packet */
      nwFreePacket(l_Packet);
    }
  }

  mmuSanityCheck();

  /* now give the packets *we* have sourced a chance */
  /* we have to re-check this because we may have just transmitted a packet */
  l_BroadcastOK=1;
  for (l_Int=p_Node->nInterface;l_Int;l_Int=l_Int->iNodeNext) {
    if (!((l_Int->iMedium->mSendPacket)(l_Int,NULL,NWADDR_INVALID))) 
      l_BroadcastOK=0;
  }

  if (l_dsr->dPackBuf) { /* we have a source packet to send */
    /* let's check our packet queue and see what's appropriate to send */
    l_SentPacket=0;
    l_dsrPacket=l_dsr->dPackBuf;
    do {
      /* Check what needs to be done with this packet */
      switch(l_dsrPacket->pStatus) {
        case DSR_STATUS_INVALID:
          /* should this ever happen? */
          printf("***ERROR*** packet %li with invalid status\n",
            l_dsrPacket->pPacket->pID);
          *((int*)(0x00))=1;
          break;
        case DSR_STATUS_VALID:
          /* check route status */
          l_Route=l_dsrPacket->pRoute;
          l_Route->rTimeUsed=p_Node->nNodeCap->nNet->nTick;
          switch(l_Route->rStatus) {
            case DSR_STATUS_INVALID:
	      if (!l_BroadcastOK)
	        break; /* we can't do anything with this right now */
              mmuSanityCheck();
              /* try a level 1 route discovery */
#ifdef DSRDEBUG
              printf("%8li DSR: Trying Level 1 route discover for packet %li\n",
                p_Node->nNodeCap->nNet->nTick,
                l_dsrPacket->pPacket->pID);
#endif
              l_Packet=nwAllocPacket();
              l_Packet->pCreateTick=p_Node->nSubNet->sNet->nTick;
              l_Packet->pSourceID=p_Node->nNodeCap->nID;
              l_Packet->pDestID=NWADDR_BROADCAST;
	      l_Packet->pNamConversationID=p_Node->nNodeCap->nID;
	      l_Packet->pNamColour=DSR_COLOUR_ROUTEQUERY1;
	      strcpy(l_Packet->pNamString,"DSR-Q1");
              /* allocate some data space for this sucker */
              l_Packet->pDataLen=sizeof(DSRPH_RQ)+sizeof(DSRPH_ROUTE);
              l_Packet->pData=mmuAlloc(l_Packet->pDataLen);
              l_dsrPH=((void*)(l_Packet->pData));
              l_dsrPH_RQ=((void*)(l_Packet->pData));
              l_Dest=((void*)(l_dsrPH))+sizeof(DSRPH_RQ);
              l_dsrPH->pType=DSR_PTYPE_ROUTEQUERY;
              l_dsrPH->pAddresses=1;
	      l_dsrPH->pHopNum=0; 
              l_Dest->pNodeCapID=p_Node->nNodeCap->nID;
#ifdef DSR_OPTION_INTERFACEROUTE
	      l_Dest->pSrcInterface=NWADDR_INVALID; /* we customize pDstInterface below */
#endif
              l_dsrPH_RQ->pSeqNum=l_dsr->dSequenceNumber;
              l_dsr->dSequenceNumber++;
              l_dsrPH_RQ->pHopLimit=DSR_ROUTE1QUERY_HOPMAX;
              l_dsrPH_RQ->pDestination=l_dsrPacket->pDestination;
              /* ok, I think we're ready to rumble */
              for (l_Int=p_Node->nInterface;l_Int;l_Int=l_Int->iNodeNext) {
#ifdef DSR_OPTION_INTERFACEROUTE
	        l_Dest->pDstInterface=l_Int->iID; /* Record which interface it went out on */
#endif
                (l_Int->iMedium->mSendPacket)(l_Int,l_Packet,NWADDR_BROADCAST);
              }
              nwFreePacket(l_Packet);
              /* now upgrade this route's status accordingly */
              l_Route->rStatus=DSR_STATUS_ROUTEDIS1;
              l_Route->rTimeout=DSR_ROUTE1QUERY_TIMEOUT;
              l_Route->rTimeoutVal=DSR_ROUTE1QUERY_TIMEOUT;
              l_SentPacket=1; /* break from our loop */
              break;
            case DSR_STATUS_ROUTEDIS1:
	      if (!l_BroadcastOK)
	        break; /* we can't do anything with this right now */
              if (l_Route->rTimeout)
                break; /* not time to resend yet */
#ifdef DSRDEBUG
              printf("%8li DSR: Trying Level 2 route discover for packet %li\n",
                p_Node->nNodeCap->nNet->nTick,
                l_dsrPacket->pPacket->pID);
#endif
              /* try a level 2 route discovery (if our Route1 has expired) */
              l_Packet=nwAllocPacket();
              l_Packet->pCreateTick=p_Node->nSubNet->sNet->nTick;
              l_Packet->pSourceID=p_Node->nNodeCap->nID;
              l_Packet->pDestID=NWADDR_BROADCAST;
	      l_Packet->pNamConversationID=p_Node->nNodeCap->nID;
	      l_Packet->pNamColour=DSR_COLOUR_ROUTEQUERY2;
	      strcpy(l_Packet->pNamString,"DSR-Q2");
              /* allocate some data space for this sucker */
              l_Packet->pDataLen=sizeof(DSRPH_RQ)+sizeof(DSRPH_ROUTE);
              l_Packet->pData=mmuAlloc(l_Packet->pDataLen);
              l_dsrPH=((void*)(l_Packet->pData));
              l_dsrPH_RQ=((void*)(l_Packet->pData));
              l_Dest=((void*)(l_dsrPH))+sizeof(DSRPH_RQ);
              l_dsrPH->pType=DSR_PTYPE_ROUTEQUERY;
              l_dsrPH->pAddresses=1;
	      l_dsrPH->pHopNum=0; 
              l_Dest->pNodeCapID=p_Node->nNodeCap->nID;
#ifdef DSR_OPTION_INTERFACEROUTE
	      l_Dest->pSrcInterface=NWADDR_INVALID; /* we customize pDstInterface below */
#endif
              l_dsrPH_RQ->pSeqNum=l_dsr->dSequenceNumber;
              l_dsr->dSequenceNumber++;
              l_dsrPH_RQ->pHopLimit=DSR_ROUTE2QUERY_HOPMAX; 
              l_dsrPH_RQ->pDestination=l_dsrPacket->pDestination;
              /* ok, I think we're ready to rumble */
              for (l_Int=p_Node->nInterface;l_Int;l_Int=l_Int->iNodeNext) {
#ifdef DSR_OPTION_INTERFACEROUTE
	        l_Dest->pDstInterface=l_Int->iID; /* Record which interface it went out on */
#endif
                (l_Int->iMedium->mSendPacket)(l_Int,l_Packet,NWADDR_BROADCAST);
              }
              nwFreePacket(l_Packet);
              /* now upgrade this route's status accordingly */
              l_Route->rStatus=DSR_STATUS_ROUTEDIS2;
              l_Route->rTimeout=DSR_ROUTE2QUERY_TIMEOUT;
              l_Route->rTimeoutVal=DSR_ROUTE2QUERY_TIMEOUT;
              l_SentPacket=1; /* break from our loop */
              break;
            case DSR_STATUS_ROUTEDIS2:
              mmuSanityCheck();
	      if (!l_BroadcastOK)
	        break; /* we can't do anything with this right now */
              /* backoff geometrically and resend route discovery */
              if (l_Route->rTimeout)
                break; /* not time to resend yet */
#ifdef DSRDEBUG
              printf("%8li DSR: Trying Level 2a route discover for packet %li\n",
                p_Node->nNodeCap->nNet->nTick,
                l_dsrPacket->pPacket->pID);
#endif
              l_Packet=nwAllocPacket();
              l_Packet->pCreateTick=p_Node->nSubNet->sNet->nTick;
              l_Packet->pSourceID=p_Node->nNodeCap->nID;
              l_Packet->pDestID=NWADDR_BROADCAST;
	      l_Packet->pNamConversationID=p_Node->nNodeCap->nID;
	      l_Packet->pNamColour=DSR_COLOUR_ROUTEQUERY2;
	      strcpy(l_Packet->pNamString,"DSR-Q2a");
              /* allocate some data space for this sucker */
              l_Packet->pDataLen=sizeof(DSRPH_RQ)+sizeof(DSRPH_ROUTE);
              l_Packet->pData=mmuAlloc(l_Packet->pDataLen);
              l_dsrPH=((void*)(l_Packet->pData));
              l_dsrPH_RQ=((void*)(l_Packet->pData));
              l_Dest=((void*)(l_dsrPH))+sizeof(DSRPH_RQ);
              l_dsrPH->pType=DSR_PTYPE_ROUTEQUERY;
              l_dsrPH->pAddresses=1;
	      l_dsrPH->pHopNum=0; 
              l_Dest->pNodeCapID=p_Node->nNodeCap->nID;
#ifdef DSR_OPTION_INTERFACEROUTE
	      l_Dest->pSrcInterface=NWADDR_INVALID; /* we customize pDstInterface below */
#endif
              l_dsrPH_RQ->pSeqNum=l_dsr->dSequenceNumber;
              l_dsr->dSequenceNumber++;
              l_dsrPH_RQ->pHopLimit=DSR_ROUTE2QUERY_HOPMAX; 
              l_dsrPH_RQ->pDestination=l_dsrPacket->pDestination;
              /* ok, I think we're ready to rumble */
              for (l_Int=p_Node->nInterface;l_Int;l_Int=l_Int->iNodeNext) {
#ifdef DSR_OPTION_INTERFACEROUTE
	        l_Dest->pDstInterface=l_Int->iID; /* Record which interface it went out on */
#endif
                (l_Int->iMedium->mSendPacket)(l_Int,l_Packet,NWADDR_BROADCAST);
              }
              nwFreePacket(l_Packet);
              /* now upgrade this route's status accordingly */
              l_Route->rStatus=DSR_STATUS_ROUTEDIS2;
              l_Route->rTimeoutVal*=DSR_ROUTE2QUERY_BACKOFF;
              if (l_Route->rTimeoutVal>DSR_ROUTE2QUERY_TIMEOUTMAX)
                l_Route->rTimeoutVal=DSR_ROUTE2QUERY_TIMEOUTMAX;
              l_Route->rTimeout=l_Route->rTimeoutVal;
              l_SentPacket=1; /* break from our loop */
              break;
            case DSR_STATUS_VALID:
              mmuSanityCheck();
              /* we have a valid route! Check if the required interface is free */
#ifdef DSR_OPTION_INTERFACEROUTE
	      l_SrcIntID=l_Route->rHop[0].pDstInterface;
	      l_DstIntID=l_Route->rHop[1].pSrcInterface;
	      if ((l_SrcIntID==NWADDR_BROADCAST)&&(!l_BroadcastOK))
	        break; /* we can't do anything with this packet right now */
	      /* check if interface is free */
              for (l_Int=p_Node->nInterface;l_Int;l_Int=l_Int->iNodeNext) {
	        if (l_Int->iID==l_SrcIntID)
		  break;
              }
              if (!(l_Int->iMedium->mSendPacket)(l_Int,NULL,NWADDR_INVALID))
	        break; /* can't transmit this just now */
#else
              l_SrcIntID=NWADDR_BROADCAST;
              l_DstIntID=NWADDR_BROADCAST; 
	      if (!l_BroadcastOK)
	        break; /* we can't do anything with this right now */
#endif
	      /* The interface is free!! Compose and send the packet. */
              /* We're actually just going to use the buffered packet 
               * so that we don't have to alloc a completely new one. */
              /*l_Packet=nwCopyPacket(l_dsrPacket);*/
              l_Packet=l_dsrPacket->pPacket;

	      /* remove this packet from the queue (NAM) - we need to 
	       * do this here because if we do it further down, the 
	       * packet will have a new pDataLen and NAM will think 
	       * it's a different packet. Thus the one in the queue will
	       * be stuck there forever :( */
	      namPacketQueueRemove(p_Node,l_Packet);
              /* We need to know how much space we need for the
               * source route. */
              for (l_i=0;
                (l_Route->rHop[l_i].pNodeCapID!=l_Packet->pDestID)&&(l_i<DSR_RCENTRY_LIMIT);
                l_i++);
              l_i++; 
              /* allocate some data space for this sucker */
              l_Packet->pDataLen=sizeof(DSRPH)+(sizeof(DSRPH_ROUTE)*l_i);
              l_Packet->pData=mmuAlloc(l_Packet->pDataLen);
              l_dsrPH=((void*)(l_Packet->pData));
              l_Dest=((void*)(l_dsrPH))+sizeof(DSRPH);
              l_dsrPH->pType=DSR_PTYPE_DATA;
              l_dsrPH->pAddresses=l_i;
	      l_dsrPH->pHopNum=0; 
	      l_Packet->pNamConversationID=p_Node->nNodeCap->nID;
	      l_Packet->pNamColour=DSR_COLOUR_DATA;
	      strcpy(l_Packet->pNamString,"DSR-D");
              /* let's collect some stats here: l_i==# source route hops */
	      /* actually, let's not. It confuses the issue with multiple interfaces and I'm getting tired of coding tonight. */
/*              p_Node->nSubNet->sPacketHop[l_i-1]++; */ /* -1 kuz first addr = hop 0 */
/*              l_k=nwNodeToInterfaceID(l_Route->rHop[l_i-1].pNodeCapID,l_Int);
	      if (l_Int->iMedium->mMinHop) {
                l_k=(l_Int->iMedium->mMinHop)(l_Int,l_k); 
                p_Node->nSubNet->sPacketHopMin[l_k]++;
	      }  
*/
              /* now fill in the source route */
              l_i=-1;
              do {
                l_i++;
		/* copy over entire DSRPH_ROUTE record */
                l_Dest[l_i]=l_Route->rHop[l_i];
              } while((l_Route->rHop[l_i].pNodeCapID!=l_Packet->pDestID)&&(l_i<DSR_RCENTRY_LIMIT));
              if (l_i>=DSR_RCENTRY_LIMIT)
                *((int*)(0x00))=8;
              /* ok, I think we're ready to rumble */
#ifdef DSRDEBUG
              printf("%8li DSR: transmitting packet %li\n",
                p_Node->nNodeCap->nNet->nTick,
                l_dsrPacket->pPacket->pID);
#endif
              l_i=l_Route->rHop[1].pNodeCapID;
              for (l_Int=p_Node->nInterface;l_Int;l_Int=l_Int->iNodeNext) {
	        if ((l_Int->iID==l_SrcIntID)||(l_SrcIntID==NWADDR_BROADCAST))
                  (l_Int->iMedium->mSendPacket)(l_Int,l_Packet,l_DstIntID);
              }
              mmuSanityCheck();
              /* we now have to remove this dsrPacket from our source buffer */
              dsrFreePacket(l_dsrPacket,p_Node);
              l_dsrPacket=NULL;
              /* now update this route */
              l_SentPacket=1; /* break from our loop */
              break;
          }
        break;
      }
      if (l_dsrPacket)
        l_dsrPacket=l_dsrPacket->pNext;
    } while((l_dsrPacket)&&(!l_SentPacket)) ;
  }
  /* we're done our tick duties! */
  mmuSanityCheck();
}

void dsrKillNode(NWNODE *p_Node) {
  DSRPACKET *l_dsrPacket,*l_dsrPacket2;
  DSRDATA *l_Dsr;
  l_Dsr=p_Node->nGeneral;

  /* first free any queued packets */
  l_dsrPacket=l_Dsr->dPackBuf;
  while(l_dsrPacket) {
    /* free this packet */
    nwFreePacket(l_dsrPacket->pPacket);
    l_dsrPacket2=l_dsrPacket;
    l_dsrPacket=l_dsrPacket->pNext;
    mmuFree(l_dsrPacket2);
  }

  /* free structure */
  mmuFree(p_Node->nGeneral);
  return;
}

void dsrRcvPacket(NWINTERFACE *p_Interface,NWPACKET *p_Packet) {
  NWPACKET *l_Packet;
  NWNODE *l_Node;
  DSRDATA *l_dsr;
  DSRPH_ROUTE *l_srcRoute,*l_dstRoute;
  DSRPH *l_dsrPH;
  DSRPH_RQ *l_dsrPH_RQ;
  DSRPH_RR *l_dsrPH_RR;
  DSRPH_RE *l_dsrPH_RE;
  long l_i,l_j;
  long l_NextHop;
  long l_Dest=0;
  long l_NumAddr;

  l_Node=p_Interface->iNode;
  l_dsr=l_Node->nGeneral;
  l_dsrPH=((DSRPH*)(p_Packet->pData));
  l_NumAddr=l_dsrPH->pAddresses;
  l_dsrPH->pHopNum++;
  l_NextHop=dsrGetNextHop(p_Packet,l_Node);

#ifdef DSRDEBUG
  printf("%8li DSR: dsrRcvPacket from Int %li with packet %li\n",
    p_Interface->iNode->nNodeCap->nNet->nTick,
    p_Interface->iID,p_Packet->pID);
#endif
  mmuSanityCheck();
  switch (l_dsrPH->pType) {
    case DSR_PTYPE_DATA:
      l_srcRoute=((void*)(l_dsrPH))+sizeof(DSRPH);
      l_Dest=l_srcRoute[l_NumAddr-1].pNodeCapID;
#ifdef DSRDEBUG
      printf("%8li DSR: node %li received DATA packet %li\n",
        p_Interface->iNode->nNodeCap->nNet->nTick,
        l_Node->nNodeCap->nID,p_Packet->pID);
#endif
      if (l_Dest==l_Node->nNodeCap->nID) { /* for this node -- process */
#ifdef DSRDEBUG
        printf("         DSR: **** Got packet %li (%li to %li)!\n",
        p_Packet->pID,p_Packet->pSourceID,p_Packet->pDestID);
#endif
        dsrStatCount(l_Node,DSR_S_PACKETS_RECEIVED);
        /* should we return a packet? */
	/* if (1)  { */ /* YES. */
	if (0) { /* No. */
          if ((p_Packet->pPayloadLen==32)||
              (p_Packet->pPayloadLen==1000)) {
            nwSourcePacket(l_Node->nNodeCap,p_Packet->pPayloadLen+1,
              p_Packet->pSourceID);
            dsrStatCount(l_Node,DSR_S_PACKETS_REPLY);
          }
	}
      } else if ((l_NextHop!=NWADDR_INVALID)) { /* in path -- forward */
        /* update route cache? */

	/* insert packet into the queue */
	namPacketQueueAdd(l_Node,p_Packet);
	/* and forward it on to its next hop */
	if (dsrRespondBufSize(l_Node)<DSR_RESPONDBUF_LIMIT) {
          l_Packet=dsrAllocRespondPacket(p_Packet,l_Node);
	} else {
	  /* buffer overflow! Discard the packet */
	  namPacketQueueDiscard(l_Node,p_Packet);
	}
      } else { /* overheard -- ignore */
        /* add route into route cache? */
      } 
      break;
    case DSR_PTYPE_ROUTEQUERY:
      l_dsrPH_RQ=((void*)(l_dsrPH));
      l_srcRoute=((void*)(l_dsrPH))+sizeof(DSRPH_RQ);
#ifdef DSRDEBUG
      printf("%8li DSR: node %li received QUERY packet %li\n",
        p_Interface->iNode->nNodeCap->nNet->nTick,
        l_Node->nNodeCap->nID,p_Packet->pID);
#endif
      /* remember to:
       *   -- add/check against query cache
       *   -- reflect known routes back to sender */
      /* we always pay attention to these -- they are broadcast */
      if (l_dsrPH_RQ->pDestination==l_Node->nNodeCap->nID) { /* for this node -- process */
        /* This packet is searching US! Let's compose a reply and buffer
         * it into our respond buffer */
        l_Packet=nwAllocPacket();
        l_Packet->pCreateTick=l_Node->nNodeCap->nNet->nTick;
        l_Packet->pSourceID=l_Node->nNodeCap->nID;
        l_Packet->pDestID=l_srcRoute[0].pNodeCapID;
	l_Packet->pNamConversationID=l_Packet->pDestID;
	l_Packet->pNamColour=DSR_COLOUR_ROUTEREPLY;
	strcpy(l_Packet->pNamString,"DSR-R");
        l_Packet->pHoldoff=
          rand()%(DSR_RRQ_HOLDOFFMAX-DSR_RRQ_HOLDOFFMIN)
          +DSR_RRQ_HOLDOFFMIN;
        /*l_Packet->pSourceIntID=l_Node->nInterface->iID;*/
        l_Packet->pDataLen=sizeof(DSRPH_RR)+sizeof(DSRPH_ROUTE)*(l_NumAddr+1);
        l_Packet->pData=mmuAlloc(l_Packet->pDataLen);
        l_dsrPH_RR=((void*)(l_Packet->pData));
        l_dstRoute=((void*)(l_dsrPH_RR))+sizeof(DSRPH_RR);
        /* copy our header info */
        ((DSRPH*)(l_dsrPH_RR))->pType=DSR_PTYPE_ROUTEREPLY;
        ((DSRPH*)(l_dsrPH_RR))->pAddresses=l_NumAddr+1;
        ((DSRPH*)(l_dsrPH_RR))->pHopNum=0;
        l_dsrPH_RR->pSeqNum=l_dsrPH_RQ->pSeqNum;
        /* copy over our source route */
        l_dstRoute[0].pNodeCapID=l_Node->nNodeCap->nID;
#ifdef DSR_OPTION_INTERFACEROUTE
        l_dstRoute[0].pSrcInterface=NWADDR_INVALID;  /* incoming - which is US so it's invalid */
        l_dstRoute[0].pDstInterface=p_Interface->iID; /* outgoing interface ID */
#endif
        for (l_i=0;l_i<l_NumAddr;l_i++) {
          l_dstRoute[l_NumAddr-l_i].pNodeCapID=l_srcRoute[l_i].pNodeCapID;
#ifdef DSR_OPTION_INTERFACEROUTE
	  /* we reverse the order of the interfaces kuz we're now going
	   * in the opposite direction!! */
          l_dstRoute[l_NumAddr-l_i].pSrcInterface=l_srcRoute[l_i].pDstInterface;
          l_dstRoute[l_NumAddr-l_i].pDstInterface=l_srcRoute[l_i].pSrcInterface;
#endif
        }
        /* record this route -- for later reference */
        dsrAddRouteSFirst(l_srcRoute[0],l_dstRoute,l_Node);
        /* ok... I think we're ready to send it. */
        namPacketQueueAdd(l_Node,l_Packet);
        if (dsrRespondBufSize(l_Node)<DSR_RESPONDBUF_LIMIT) {
          /* and add the sucker to our buffer! */
          dsrInsertRespondPacket(l_Packet, l_Node);
        } else {
          /* no room in buffer - drop it */
          namPacketQueueDiscard(l_Node,l_Packet);
          nwFreePacket(l_Packet);
        }
        /* and we're done! */
      /*} else if (dsrFindRoute(l_Dest,l_Node)) {*/
        /* compose reply with the route from our cache */
/* #define DSR_RRQ_HOPMULT*/
      } else {
        /* we can't help -- forward the sucker if hop limit not exceeded 
         * and if we're not already listed in the route.                 */
        /* Unfortunately, we have to compose a NEW packet with our address
         * appended on the source route */
        if ( (l_NextHop==NWADDR_INVALID)&&(l_dsrPH_RQ->pHopLimit>l_NumAddr)&&
             (dsrFindNodeInRoute(p_Packet,l_Node)==-1) ) {
          /* first, check if this source-sequence pair are in our cache */
          for (l_i=0;l_i<DSR_RRCACHE_LIMIT;l_i++) {
            if ((l_dsr->dRRCache[l_i].sSource==l_srcRoute[0].pNodeCapID) &&
                (l_dsr->dRRCache[l_i].sSequenceNumber==l_dsrPH_RQ->pSeqNum)) {
              break;
            } /* else we haven't seen the packet yet */
          }
          if (l_i==DSR_RRCACHE_LIMIT) {
#ifdef DSRDEBUG
            printf("%8li DSR: forwarding QUERY packet %li\n",
              p_Interface->iNode->nNodeCap->nNet->nTick,
              p_Packet->pID);
#endif
            /* add to our cache */
            l_j=0;
            for (l_i=1;l_i<DSR_RRCACHE_LIMIT;l_i++) {
              if ((l_dsr->dRRCache[l_i].sTimeStamp<l_dsr->dRRCache[l_j].sTimeStamp)||
                  (l_dsr->dRRCache[l_i].sSource==NWADDR_INVALID)) {
                l_j=l_i;
              }
            }
            l_dsr->dRRCache[l_j].sSource=l_srcRoute[0].pNodeCapID;
            l_dsr->dRRCache[l_j].sDestination=l_Dest;
            l_dsr->dRRCache[l_j].sSequenceNumber=l_dsrPH_RQ->pSeqNum;
            l_dsr->dRRCache[l_j].sTimeStamp=l_Node->nNodeCap->nNet->nTick;
            /* forward packet */
            l_Packet=nwAllocPacket();
            l_Packet->pID=p_Packet->pID;
            l_Packet->pCreateTick=p_Packet->pCreateTick;
            l_Packet->pSourceID=p_Packet->pSourceID;
            l_Packet->pDestID=p_Packet->pDestID;
	    l_Packet->pNamConversationID=l_Packet->pSourceID;
	    l_Packet->pNamColour=DSR_COLOUR_ROUTEQUERY;
	    strcpy(l_Packet->pNamString,"DSR-Q");
            l_Packet->pHops=p_Packet->pHops+1;
            l_Packet->pTimeout=0L;
            l_Packet->pHoldoff=
              rand()%(DSR_FORWARD_HOLDOFFMAX-DSR_FORWARD_HOLDOFFMIN)
              +DSR_FORWARD_HOLDOFFMIN;
            l_Packet->pDataLen=sizeof(DSRPH_RQ)+sizeof(DSRPH_ROUTE)*(l_NumAddr+1);
            l_Packet->pData=mmuAlloc(l_Packet->pDataLen);
            ((DSRPH*)(l_Packet->pData))->pType=l_dsrPH->pType;
            ((DSRPH*)(l_Packet->pData))->pAddresses=l_NumAddr+1;
            ((DSRPH*)(l_Packet->pData))->pHopNum=l_dsrPH->pHopNum;
            ((DSRPH_RQ*)(l_Packet->pData))->pSeqNum=l_dsrPH_RQ->pSeqNum;
            ((DSRPH_RQ*)(l_Packet->pData))->pHopLimit=l_dsrPH_RQ->pHopLimit;
            ((DSRPH_RQ*)(l_Packet->pData))->pDestination=l_dsrPH_RQ->pDestination;
            l_dstRoute=((void*)(l_Packet->pData))+sizeof(DSRPH_RQ);
            for (l_i=0;l_i<l_NumAddr;l_i++) {
	      /* the following line will copy ALL DSRPH_ROUTE fields */
              l_dstRoute[l_i]=l_srcRoute[l_i];
            }
	    /* now add US into the list */
            l_dstRoute[l_i].pNodeCapID=l_Node->nNodeCap->nID;
#ifdef DSR_OPTION_INTERFACEROUTE
	    l_dstRoute[l_i].pSrcInterface=p_Interface->iID; /* where the pkt came from */
	    l_dstRoute[l_i].pDstInterface=NWADDR_BROADCAST; /* where it's going */
	      /* - note when we actually send the packet, we'll come
	       * back to this field and put the actual interface ID in 
	       * pDstInterface as we pass the packet to each interface */
#endif
            /* ok... I think we're ready to send it. */
            namPacketQueueAdd(l_Node,l_Packet);
            if (dsrRespondBufSize(l_Node)<DSR_RESPONDBUF_LIMIT) {
              /* ok... I think we're ready to send it. */
              dsrInsertRespondPacket(l_Packet, l_Node);
            } else {
              /* no room in buffer - drop it */
              namPacketQueueDiscard(l_Node,l_Packet);
              nwFreePacket(l_Packet);
            }
          } /* else we've seen this request before -- ignore */
        } /* else ignore */
      }
      break;
    case DSR_PTYPE_ROUTEREPLY: 
      l_dsrPH_RR=((void*)(l_dsrPH));
      l_srcRoute=((void*)(l_dsrPH))+sizeof(DSRPH_RR);
      l_Dest=l_srcRoute[l_NumAddr-1].pNodeCapID;
#ifdef DSRDEBUG
      printf("%8li DSR: node %li received REPLY packet %li bound for %li\n",
        p_Interface->iNode->nNodeCap->nNet->nTick,
        l_Node->nNodeCap->nID,p_Packet->pID,l_Dest);
#endif
      if (l_Dest==l_Node->nNodeCap->nID) { /* for this node -- process */
        DSRROUTE *l_tmpRoute;
        l_tmpRoute=dsrFindRoute(*l_srcRoute,l_Node);
        if((!l_tmpRoute)||((l_tmpRoute)&&(l_tmpRoute->rStatus!=DSR_STATUS_VALID))) {
          printf("%8li.%04li DSR: node %li route requirement satisfied.\n",
            p_Interface->iNode->nNodeCap->nNet->nTick/NWTICK,
            p_Interface->iNode->nNodeCap->nNet->nTick%NWTICK,
            l_Node->nNodeCap->nID);
          dsrAddRouteDFirst(l_srcRoute,l_Node);
          if (dsrCheckForConvergence(p_Interface->iNode)&&(p_Interface->iNode->nSubNet->sReportConvergence)) {
            printf("%8li.%04li DSR: Subnetwork has converged.\n",
              p_Interface->iNode->nNodeCap->nNet->nTick/NWTICK,
              p_Interface->iNode->nNodeCap->nNet->nTick%NWTICK);
          }
        }

	/*
	else {
          printf("%8li DSR: node %li route duplication\n",
            p_Interface->iNode->nNodeCap->nNet->nTick/(NWTICK/1000),
            l_Node->nNodeCap->nID);
	  
	}
	*/

      } else if ((l_NextHop!=NWADDR_INVALID)) { /* in path -- forward */
	/* insert packet into the queue */
	namPacketQueueAdd(l_Node,p_Packet);
        /* forward this packet */
	if (dsrRespondBufSize(l_Node)<DSR_RESPONDBUF_LIMIT) {
          l_Packet=dsrAllocRespondPacket(p_Packet,l_Node);
	} else {
	  /* buffer overflow! Discard the packet */
	  namPacketQueueDiscard(l_Node,p_Packet);
	}
      } else { /* overheard -- ignore */
        /* ignore */
      } 
      break;
    case DSR_PTYPE_ROUTEERROR:
      l_dsrPH_RE=((void*)(l_dsrPH));
      l_srcRoute=((void*)(l_dsrPH))+sizeof(DSRPH_RE);
      l_Dest=l_srcRoute[l_NumAddr-1].pNodeCapID;
#ifdef DSRDEBUG
      printf("%8li DSR: node %li received ERROR packet %li\n",
        p_Interface->iNode->nNodeCap->nNet->nTick,
        l_Node->nNodeCap->nID,p_Packet->pID);
#endif
      dsrRemoveRoute(l_dsrPH_RE->pSender,l_dsrPH_RE->pReceiver,l_Node);
      if (l_Dest==l_Node->nNodeCap->nID) { /* for this node -- process */
        /* we've already updated our route -- ignore */
      } else if ((l_NextHop!=NWADDR_INVALID)) { /* in path -- forward */
	/* insert packet into the queue */
	namPacketQueueAdd(l_Node,p_Packet);
	if (dsrRespondBufSize(l_Node)<DSR_RESPONDBUF_LIMIT) {
          l_Packet=dsrAllocRespondPacket(p_Packet,l_Node);
	} else {
	  /* buffer overflow! Discard the packet */
	  namPacketQueueDiscard(l_Node,p_Packet);
	}
      } else { /* overheard -- ignore */
        /* we've already updated our route -- ignore */
      } 
      break;
    default:
      printf("****ERROR**** invalid DSR packet type\n");
      *((int*)(0x00))=2;
      break;
  }
  mmuSanityCheck();

/*
  printf("%8li DSR: RcvPacket from node %li to node %li(for %li), via medium %li\n",
    p_Interface->iNode->nNodeCap->nNet->nTick,
    p_Packet->pSourceID,
    p_Interface->iNode->nNodeCap->nID,
    p_Packet->pDestID,
    p_Interface->iMedium->mID);
*/ 
}

void dsrDeliveryError(NWINTERFACE *p_Interface,NWPACKET *p_Packet, long p_ErrorCode) {
  DSRDATA *l_dsr;
  DSRPH *l_dsrPH;
  NWNODE *l_Node;
  DSRPH_ROUTE *l_srcRoute,*l_dstRoute;
  NWPACKET *l_Packet;
  long l_i;
  long l_NumAddr;
  long l_NextHop;

  l_Node=p_Interface->iNode;
  l_dsr=l_Node->nGeneral;
  l_dsrPH=((DSRPH*)(p_Packet->pData));

  l_srcRoute=dsrIndexToRoute(p_Packet);
  
  l_NextHop=dsrGetNextHop(p_Packet,l_Node);

  if ((l_NextHop==NWADDR_INVALID) || (l_NextHop==NWADDR_BROADCAST)) {
    *((int*)(0x00))=1;
    return;
  }

  mmuSanityCheck();
  
  /* we have to do 2 things: 
   *   1) remove this link from any routes in our cache 
   *   2) if we are forwarding this packet, return an ERROR packet to the sender.
   */
  dsrRemoveRoute(l_Node->nNodeCap->nID,l_NextHop,l_Node);

  /* ok, now let's make packets -- if we're not the originator, that is!!!! */
  if (l_srcRoute[0].pNodeCapID!=l_Node->nNodeCap->nID) { 
    /* first, find us in the source route */
    for (l_NumAddr=0;
      (l_srcRoute[l_NumAddr].pNodeCapID!=l_Node->nNodeCap->nID)&&(l_NumAddr<DSR_RCENTRY_LIMIT);
      l_NumAddr++);

    /* we need to send a packet */
    l_Packet=nwAllocPacket();
    l_Packet->pCreateTick=p_Packet->pCreateTick;
    l_Packet->pSourceID=l_Node->nNodeCap->nID;
    l_Packet->pDestID=l_srcRoute[0].pNodeCapID;
    l_Packet->pNamConversationID=l_Packet->pDestID;
    l_Packet->pNamColour=DSR_COLOUR_ROUTEERROR;
    strcpy(l_Packet->pNamString,"DSR-E");
    l_Packet->pHops=0;
    l_Packet->pTimeout=0L;
    l_Packet->pHoldoff=
      rand()%(DSR_FORWARD_HOLDOFFMAX-DSR_FORWARD_HOLDOFFMIN)
      +DSR_FORWARD_HOLDOFFMIN;
    l_Packet->pDataLen=sizeof(DSRPH_RE)+sizeof(DSRPH_ROUTE)*(l_NumAddr+1);
    l_Packet->pData=mmuAlloc(l_Packet->pDataLen);
    ((DSRPH*)(l_Packet->pData))->pType=DSR_PTYPE_ROUTEERROR;
    ((DSRPH*)(l_Packet->pData))->pAddresses=l_NumAddr+1;
    ((DSRPH*)(l_Packet->pData))->pHopNum=0;
    ((DSRPH_RE*)(l_Packet->pData))->pDestination=l_srcRoute[0].pNodeCapID;
    ((DSRPH_RE*)(l_Packet->pData))->pSender=l_Node->nNodeCap->nID;
    ((DSRPH_RE*)(l_Packet->pData))->pReceiver=l_NextHop;
    l_dstRoute=((void*)(l_Packet->pData))+sizeof(DSRPH_RE);
    for (l_i=0;l_i<(l_NumAddr+1);l_i++) {
      l_dstRoute[l_NumAddr-l_i].pNodeCapID=l_srcRoute[l_i].pNodeCapID;
#ifdef DSR_OPTION_INTERFACEROUTE
      l_dstRoute[l_NumAddr-l_i].pSrcInterface=l_srcRoute[l_i].pDstInterface;
      l_dstRoute[l_NumAddr-l_i].pDstInterface=l_srcRoute[l_i].pSrcInterface;
#endif
    }
    /* ok... I think we're ready to send it. */
    namPacketQueueAdd(l_Node,l_Packet);
    if (dsrRespondBufSize(l_Node)<DSR_RESPONDBUF_LIMIT) {
      dsrInsertRespondPacket(l_Packet, l_Node);
    } else {
      /* no room in buffer - drop it */
      namPacketQueueDiscard(l_Node,l_Packet);
      nwFreePacket(l_Packet);
    }
  }

#ifdef DSRDEBUG
  printf("%8li DSR: DELIVERY FAILED from node %li to node %li, via Int %li to %li, via medium %li\n",
    p_Interface->iNode->nNodeCap->nNet->nTick,
    p_Packet->pSourceID,
    p_Packet->pDestID,
    p_Packet->pSourceIntID,
    p_Packet->pDestIntID,
    p_Interface->iMedium->mID);
#endif
  /* resubmit packet in broadcast mode */
/*  (p_Interface->iMedium->mSendPacket)(p_Interface,p_Packet,NWADDR_BROADCAST);*/
}

