/* network.c 
 * Two-letter symbol: nw
 * This module implements the network simulation environment
 * and provides calls to construct networks, operate them, and
 * push data around. 
 * NOTE: This module is meant to be a "simulation management" module.
 * Therefore, different protocols at network and data link layers
 * should be implemented elsewhere and called by this module.
 */

#include<libio.h>
#include<string.h>
#include"network.h"
#include"memmgr.h"
#include"schedule.h"
#include"nam.h"

/* Globals */
long g_nwNodeCapIDPool=1000L;
long g_nwPacketIDPool=5000000L;
long g_nwLastTickPacketID=0;
long g_nwMediumIDPool=2000L;
long g_nwInterfaceIDPool=3000L;

/* Set up by this proc: nID,  nPrev,nNext,nNet (if p_Net provided)
 * You have to set up nX, nY, nInitNodeCap, nInitTickNodeCap, nKillNodeCap.  */
NWNODECAP *nwAllocNodeCap(NWNETWORK *p_Net) {
  NWNODECAP *l_NodeCap;
  int l_i;

  l_NodeCap=mmuAlloc(sizeof(NWNODECAP));
  if (l_NodeCap) {
    l_NodeCap->nID=g_nwNodeCapIDPool;
    g_nwNodeCapIDPool++;
    l_NodeCap->nX=l_NodeCap->nY=0;
    for (l_i=0;l_i<10;l_i++)
      l_NodeCap->nParam[l_i]=0;
    l_NodeCap->nInitNodeCap=NULL;
    l_NodeCap->nTickNodeCap=NULL;
    l_NodeCap->nKillNodeCap=NULL;
    l_NodeCap->nPacketBuffer=NULL;
    l_NodeCap->nNet=p_Net;
    if (p_Net) {
      /* add this sucker into the linked network list */
      l_NodeCap->nPrev=NULL;
      l_NodeCap->nNext=p_Net->nNodeCap;
      if (p_Net->nNodeCap)
        p_Net->nNodeCap->nPrev=l_NodeCap;
      p_Net->nNodeCap=l_NodeCap;
    } else {
      l_NodeCap->nPrev=l_NodeCap->nNext=NULL;
    }
    l_NodeCap->nGeneral=NULL;
  }
  return l_NodeCap;
}

/* This proc does:
 * Removes pNodeCap from its linked list. */
void nwFreeNodeCap(NWNODECAP *p_NodeCap) {
  if (!p_NodeCap)
    return;

  /* zero'th, free up any memory associated with nGeneral */
  if (p_NodeCap->nKillNodeCap)
    (p_NodeCap->nKillNodeCap)(p_NodeCap);

  /* First, remove it from its list */
  if (p_NodeCap->nNext) {
    p_NodeCap->nNext->nPrev=p_NodeCap->nPrev;
  }
  if (p_NodeCap->nPrev) {
    p_NodeCap->nPrev->nNext=p_NodeCap->nNext;
  } else {
    if (p_NodeCap->nNet) {
      p_NodeCap->nNet->nNodeCap=p_NodeCap->nNext;
    }  /* else they are going to lose this pointer! Too bad. */
  }

  /* then we just free the sucker */
  mmuFree(p_NodeCap);
}

/* fills nSubNet, nPrev, nNext.
 * You have to fill: nNodeCap, nInterface, nInitNode, nTickNode, 
 *   nKillNode, nRcvPacket, nDeliveryError. */
NWNODE *nwAllocNode (NWSUBNETWORK *p_SubNet) {
  NWNODE *l_Node;
  l_Node=mmuAlloc(sizeof(NWNODE));
  if (l_Node) {
    l_Node->nNodeCap=NULL;
    l_Node->nSubNet=p_SubNet;
    if (p_SubNet) {
      l_Node->nNext=p_SubNet->sNode;
      l_Node->nPrev=NULL;
      if (p_SubNet->sNode) 
        p_SubNet->sNode->nPrev=l_Node;
      p_SubNet->sNode=l_Node;
    } else {
      l_Node->nPrev=l_Node->nNext=NULL;
    }
    l_Node->nInterface=NULL;
    l_Node->nInitNode=NULL;
    l_Node->nTickNode=NULL;
    l_Node->nKillNode=NULL;
    l_Node->nRcvPacket=NULL;
    l_Node->nDeliveryError=NULL;
    l_Node->nGeneral=NULL;
  }
  return l_Node; 
}

void nwFreeNode(NWNODE *p_Node) {
  if (p_Node) {

    /* zero'th, kill any nGeneral memory */
    if (p_Node->nKillNode)
      (p_Node->nKillNode)(p_Node);

    /* first kill any interface this node may have attached */
    while (p_Node->nInterface)
      nwFreeInterface(p_Node->nInterface);
    /* next, extract it from its linked list */
    if (p_Node->nPrev) {
      /* it's in the middle someplace */
      p_Node->nPrev->nNext=p_Node->nNext;
    } else {
      if (p_Node->nSubNet) 
        p_Node->nSubNet->sNode=p_Node->nNext;
    }
    if (p_Node->nNext)
      p_Node->nNext->nPrev=p_Node->nPrev;

    /* lastly, free the sucker */
    mmuFree(p_Node);
  }
}

/* You have to fill in any medium-dependend info. */
/* Use the tInitMedium function in the p_Medium structure to init tech-specific
 * extra info allocated in the NWINTERFACE superstructure. */
NWINTERFACE *nwAllocInterface(NWNODE *p_Node,NWMEDIUM *p_Medium) {
  NWINTERFACE *l_Interface;

  if ((!p_Medium)||(!p_Node))
    return NULL;

  l_Interface=mmuAlloc(sizeof(NWINTERFACE));
  if (l_Interface) {
    l_Interface->iID=g_nwInterfaceIDPool;
    g_nwInterfaceIDPool++;
    l_Interface->iNode=p_Node;
    if (p_Node) {
      l_Interface->iNodeNext=p_Node->nInterface;
      l_Interface->iNodePrev=NULL;
      if (p_Node->nInterface)
        p_Node->nInterface->iNodePrev=l_Interface;
      p_Node->nInterface=l_Interface;
    } else {
      l_Interface->iNodePrev=l_Interface->iNodeNext=NULL;
    }
    l_Interface->iMedium=p_Medium;
    if (p_Medium) {
      l_Interface->iMediumNext=p_Medium->mInterface;
      l_Interface->iMediumPrev=NULL;
      if (p_Medium->mInterface)
        p_Medium->mInterface->iMediumPrev=l_Interface;
      p_Medium->mInterface=l_Interface;
    } else {
      l_Interface->iMediumPrev=l_Interface->iMediumNext=NULL;
    }
  }
  return l_Interface;
}

void nwFreeInterface(NWINTERFACE *p_Interface) {
  if (p_Interface) {

    /* zero'th, kill iGeneral */
    if (p_Interface->iMedium)
      if (p_Interface->iMedium->mKillInterface)
        (p_Interface->iMedium->mKillInterface)(p_Interface);

    /* extract from its node list */
    if (p_Interface->iNodePrev) {
      p_Interface->iNodePrev->iNodeNext=p_Interface->iNodeNext;
    } else {
      if (p_Interface->iNode)
        p_Interface->iNode->nInterface=p_Interface->iNodeNext;
    }
    if (p_Interface->iNodeNext)
      p_Interface->iNodeNext->iNodePrev=p_Interface->iNodePrev;

    /* extract from its medium list */
    if (p_Interface->iMediumPrev) {
      p_Interface->iMediumPrev->iMediumNext=p_Interface->iMediumNext;
    } else {
      if (p_Interface->iMedium)
        p_Interface->iMedium->mInterface=p_Interface->iMediumNext;
    }
    if (p_Interface->iMediumNext)
      p_Interface->iMediumNext->iMediumPrev=p_Interface->iMediumPrev;

    /* Free memory */
    mmuFree(p_Interface);
  }
}

void nwBaudToMultIncr(NWMEDIUM *p_Medium) {
  float l_float;
  long l_i;

  l_float=p_Medium->mBaud;
  l_float/=8;          /* 8 bits per byte */
  l_float/=NWTICK;     /* our simulation resolution */

  /* this just increases our resolution when Baud ~= NWTICK */
  l_i=1;
  if ((l_float<20)&&(l_float>0.05))
    l_i=10;
  if ((l_float<1.5)&&(l_float>0.75))
    l_i=100;

  if (l_float>1) {
    p_Medium->mBaudMultiplier=l_i;
    p_Medium->mBaudIncrementer=(l_i*l_float+0.5); /* 0.5 is to round up if fraction is >0.5 */
  } else {
    p_Medium->mBaudIncrementer=l_i;
    p_Medium->mBaudMultiplier=(l_i/l_float+0.5);
  }
}

void nwMultIncrToBaud(NWMEDIUM *p_Medium) { 
  /* I broke it up this way just so I could, for my own 
   * satisfaction, ensure the calculation is done in floating point */
  p_Medium->mBaud=p_Medium->mBaudIncrementer;
  p_Medium->mBaud*=8;                          /* 8 bits per byte */
  p_Medium->mBaud/=p_Medium->mBaudMultiplier;
  p_Medium->mBaud/=NWTICK;
}

float nwTicksPerByte(NWMEDIUM *p_Medium) {
  float l_float;
  if (!p_Medium) 
    return 1.0;
  l_float=NWTICK;
  l_float*=10.0;                /* 8 bits / byte */
  l_float/=p_Medium->mBaud;
  return l_float;
}

/* Done for you: mID, mSubNetwork, mPrev,mNext. */
/* You have to fill in mInterface (done by nwInterfaceAlloc), 
 * m?Medium fn's, mGeneral. */
NWMEDIUM *nwAllocMedium(NWSUBNETWORK *p_SubNet) {
  NWMEDIUM *l_Medium;
  l_Medium=mmuAlloc(sizeof(NWMEDIUM));
  if (l_Medium) {
    l_Medium->mID=g_nwMediumIDPool;
    g_nwMediumIDPool++;
    l_Medium->mName[0]=0;
    l_Medium->mSubNetwork=p_SubNet;
    l_Medium->mInterface=NULL;
    if (p_SubNet) {
      l_Medium->mNext=p_SubNet->sMedium;
      l_Medium->mPrev=NULL;
      if (p_SubNet->sMedium)
        p_SubNet->sMedium->mPrev=l_Medium;
      p_SubNet->sMedium=l_Medium;
    } else {
      l_Medium->mPrev=l_Medium->mNext=NULL;
    }
    l_Medium->mState=NWSTATE_OK;
    l_Medium->mBaudMultiplier=1;
    l_Medium->mBaudIncrementer=10;
    nwMultIncrToBaud(l_Medium);
    l_Medium->mInitMedium=NULL;
    l_Medium->mTickMedium=NULL;
    l_Medium->mKillMedium=NULL;
    l_Medium->mKillInterface=NULL;
    l_Medium->mMinHop=NULL;
    l_Medium->mSendPacket=NULL;
  } 
  return l_Medium;
}

void nwFreeMedium(NWMEDIUM *p_Medium) {
  if (p_Medium) {

    /* zero'th, free mGeneral and iGeneral */
    if (p_Medium->mKillMedium) 
      (p_Medium->mKillMedium)(p_Medium);

    /* remove from its linked list */
    if (p_Medium->mNext)
      p_Medium->mNext->mPrev=p_Medium->mPrev;
    if (p_Medium->mPrev) {
      p_Medium->mPrev=p_Medium->mNext;
    } else {
      if (p_Medium->mSubNetwork)
        p_Medium->mSubNetwork->sMedium=p_Medium->mNext;
    }

    /* Free memory */
    mmuFree(p_Medium);
  }
}

/* Fills in everything for you via other functions 
 * except for sName if you want one */
NWSUBNETWORK *nwAllocSubNetwork(NWNETWORK *p_Net) {
  NWSUBNETWORK *l_SubNet;

  l_SubNet=mmuAlloc(sizeof(NWSUBNETWORK));
  if (l_SubNet) {
    l_SubNet->sName[0]=0;
    l_SubNet->sFile=-1;
    l_SubNet->sNet=p_Net;
    l_SubNet->sNode=NULL;
    l_SubNet->sMedium=NULL;
    l_SubNet->sReportConvergence=1; /* default to ON for convergence reporting */
    if (p_Net) {
      l_SubNet->sPrev=NULL;
      l_SubNet->sNext=p_Net->nSubNetwork;
      if (p_Net->nSubNetwork)
        p_Net->nSubNetwork->sPrev=l_SubNet;
      p_Net->nSubNetwork=l_SubNet;
    } else {
      l_SubNet->sPrev=l_SubNet->sNext=NULL;
    }

  }
  return l_SubNet;
}

void nwFreeSubNetwork(NWSUBNETWORK *p_SubNet) {

  if (!p_SubNet)
    return;

  /* First free any nNode list */
  while(p_SubNet->sNode)
    nwFreeNode(p_SubNet->sNode);

  /* and any interface list */
  while(p_SubNet->sMedium)
    nwFreeMedium(p_SubNet->sMedium);

  /* next remove it from its linked list */
  if (p_SubNet->sPrev) {
    p_SubNet->sPrev->sNext=p_SubNet->sNext;
  } else {
    if (p_SubNet->sNet) {
      p_SubNet->sNet->nSubNetwork=p_SubNet->sNext;
    }
  }
  if (p_SubNet->sNext) 
    p_SubNet->sNext->sPrev=p_SubNet->sPrev;

  /* and now free it */
  mmuFree(p_SubNet);
}

NWNETWORK *nwAllocNetwork(void) {
  NWNETWORK *l_Net;
  l_Net=mmuAlloc(sizeof(NWNETWORK));
  if (l_Net) {
    l_Net->nTick=0L;
    l_Net->nNodeCap=NULL;
    l_Net->nSubNetwork=NULL;
    l_Net->nPauseTime=NW_S(4000);
    l_Net->nStopTime=NW_S(4000);
  }
  return l_Net;
}

void nwFreeNetwork(NWNETWORK *p_Net) {
  if (p_Net) {
    /* first free any subnetworks */
    while (p_Net->nSubNetwork)
      nwFreeSubNetwork(p_Net->nSubNetwork);
    /* next free nodecap structures */
    while (p_Net->nNodeCap)
      nwFreeNodeCap(p_Net->nNodeCap);
    /* last, free the structure */
    mmuFree(p_Net);
  }
}

/* you have to fill in pretty much everything except the pID. */
NWPACKET *nwAllocPacket(void) {
  NWPACKET *l_Packet;
  l_Packet=mmuAlloc(sizeof(NWPACKET));
  if (l_Packet) {
    l_Packet->pID=g_nwPacketIDPool;
    g_nwPacketIDPool++;
    l_Packet->pCreateTick=0L;
    l_Packet->pSourceID=l_Packet->pDestID=NWADDR_INVALID;
    l_Packet->pHops=0L;
    l_Packet->pPrev=l_Packet->pNext=NULL;
    l_Packet->pTimeout=0;
    l_Packet->pHoldoff=0;
    l_Packet->pSourceIntID=NWADDR_INVALID;
    l_Packet->pDestIntID=NWADDR_INVALID;
    l_Packet->pDataLen=0L;
    l_Packet->pData=NULL;
    l_Packet->pPayloadLen=0L;
  }
  return l_Packet;
}

void nwFreePacket(NWPACKET *p_Packet) {
  if (p_Packet) {
    if (p_Packet->pData)
      mmuFree(p_Packet->pData);
    mmuFree(p_Packet);
  }
}


/* network operators */
int nwInitNetwork(NWNETWORK *p_Net) {
  NWSUBNETWORK *l_SubNet;
  NWNODE *l_Node;
  NWNODECAP *l_NodeCap;
  NWMEDIUM *l_Medium;

  if (p_Net) {
    p_Net->nTick=0;
    /* init our nodecaps */
    for (l_NodeCap=p_Net->nNodeCap;l_NodeCap;l_NodeCap=l_NodeCap->nNext) {
      if (l_NodeCap->nInitNodeCap) 
        (l_NodeCap->nInitNodeCap)(l_NodeCap);
    }
    /* Initialize all our subnetworks etc */
    for (l_SubNet=p_Net->nSubNetwork;l_SubNet;l_SubNet=l_SubNet->sNext) {
      /* let's initialize all our nodes and interfaces */
      for (l_Node=l_SubNet->sNode;l_Node;l_Node=l_Node->nNext) {
        if (l_Node->nInitNode)
          (l_Node->nInitNode)(l_Node);
      }
      for (l_Medium=l_SubNet->sMedium;l_Medium;l_Medium=l_Medium->mNext) {
        if (l_Medium->mInitMedium)
          (l_Medium->mInitMedium)(l_Medium);
      }
    }
    return 1;
  }
  return 0;
}

void nwTickNetwork(NWNETWORK *p_Net) {
  NWSUBNETWORK *l_SubNet;
  NWNODE *l_Node;
  NWMEDIUM *l_Medium;
  NWNODECAP *l_NodeCap;
  NWPACKET *l_P1,*l_P2;

  if (p_Net) {
    /* very first, give our event scheduler a wack at things */
    /* Note the scheduler is designed to NOT get called 
     * every single tick, so you can scale it back if required */
    schScheduleTick(p_Net->nTick);
    

    /* first, let's give our nodecaps a wack at it */
    /* NOTE: Node tick fn's should do things like:
        a) move nodes, if appropriate
        b) generate packets 
    */
    for (l_NodeCap=p_Net->nNodeCap;l_NodeCap;l_NodeCap=l_NodeCap->nNext) {
      /* First, free any old packets in the application send buffer */
      l_P1=l_NodeCap->nPacketBuffer;
      while(l_P1) {
        l_P2=l_P1->pNext;
        if (l_P1->pID<g_nwLastTickPacketID) {
          /* free this packet */
          if (l_P1->pPrev) {
            l_P1->pPrev->pNext=l_P2;
          } else {
            l_NodeCap->nPacketBuffer=l_P2;
          }
          if (l_P2)
            l_P2->pPrev=l_P1->pPrev;
          nwFreePacket(l_P1);
        }
        l_P1=l_P2;
      }
      /* next do our tick! */
      if (l_NodeCap->nTickNodeCap)
        (l_NodeCap->nTickNodeCap)(l_NodeCap);
    }
    g_nwLastTickPacketID=g_nwPacketIDPool;

    /* Next, let's allow our network layer to do it's thing */
    for (l_SubNet=p_Net->nSubNetwork;l_SubNet;l_SubNet=l_SubNet->sNext) {
      for (l_Node=l_SubNet->sNode;l_Node;l_Node=l_Node->nNext) {
        if (l_Node->nTickNode) 
          (l_Node->nTickNode)(l_Node);
      }
    }

    /* finally, let's allow the physical layer it's time in the spotlight */
    for (l_SubNet=p_Net->nSubNetwork;l_SubNet;l_SubNet=l_SubNet->sNext) {
      for (l_Medium=l_SubNet->sMedium;l_Medium;l_Medium=l_Medium->mNext) {
        if (l_Medium->mTickMedium)
          (l_Medium->mTickMedium)(l_Medium);
      }
    }

    /* lastly, increment our tick */
    p_Net->nTick++;
  }
}

/* Application layer should call this function to generate a packet. */
/* This function will create a packet and put it in the nodecap.     */
/* NOTE: p_DestinationID is the NODECAP ID.                          */
NWPACKET *nwSourcePacket(NWNODECAP *p_NodeCap, unsigned long p_PayloadLen,long p_DestinationID)  {
  NWPACKET *l_Packet;

  if (!p_NodeCap)
    return NULL;

  l_Packet=nwAllocPacket();
  if (l_Packet) {
    l_Packet->pCreateTick=p_NodeCap->nNet->nTick;
    l_Packet->pSourceID=p_NodeCap->nID;
    l_Packet->pDestID=p_DestinationID;
    l_Packet->pHops=0;
    l_Packet->pDataLen=0;
    l_Packet->pData=NULL;
    l_Packet->pPayloadLen=p_PayloadLen;
    l_Packet->pNamConversationID=0;
    l_Packet->pNamColour=0;
    l_Packet->pNamString[0]=0;

    l_Packet->pPrev=NULL;
    l_Packet->pNext=p_NodeCap->nPacketBuffer;
    if (p_NodeCap->nPacketBuffer)  {
      p_NodeCap->nPacketBuffer->pPrev=l_Packet;
    } 
    p_NodeCap->nPacketBuffer=l_Packet;
  }
  return l_Packet;
}

/* Network/physical layer should call this function to create a copy of */
/* a packet for itself from an original.                                */
/* NOTE: pData and pDataLen are duplicated if p_CopyData is non zero;   */
/*       otherwise they are not copied or otherwise manipulated.        */
NWPACKET *nwCopyPacket(NWPACKET *p_Packet,int p_CopyData) {
  NWPACKET *l_newPacket;
  unsigned char *l_sd,*l_dd;
  int i;

  if (p_Packet) {
    l_newPacket=nwAllocPacket();
    if (l_newPacket) {
      l_newPacket->pID=p_Packet->pID;
      l_newPacket->pCreateTick=p_Packet->pCreateTick;
      l_newPacket->pSourceID=p_Packet->pSourceID;
      l_newPacket->pDestID=p_Packet->pDestID;
      l_newPacket->pHops=p_Packet->pHops;
      l_newPacket->pPrev=l_newPacket->pNext=NULL;
      l_newPacket->pSourceIntID=p_Packet->pSourceIntID;
      l_newPacket->pDestIntID=p_Packet->pDestIntID;
      l_newPacket->pNamConversationID=p_Packet->pNamConversationID;
      l_newPacket->pNamColour=p_Packet->pNamColour;
      strcpy(l_newPacket->pNamString,p_Packet->pNamString);

      if (p_CopyData && p_Packet->pData) {
        l_newPacket->pData=mmuAlloc(p_Packet->pDataLen);
        if (l_newPacket->pData) {
          l_newPacket->pDataLen=p_Packet->pDataLen;
          l_sd=p_Packet->pData;
          l_dd=l_newPacket->pData;
          for (i=0;i<p_Packet->pDataLen;i++) {
            *l_dd=*l_sd;
            l_dd++;
            l_sd++;
          }
        } else {
          l_newPacket->pDataLen=0L;
          l_newPacket->pData=NULL;
        }
      } else {
        l_newPacket->pDataLen=0L;
        l_newPacket->pData=NULL;
      }
      l_newPacket->pPayloadLen=p_Packet->pPayloadLen;
    }
    return l_newPacket;
  }
  return NULL;
}

/* This proc is essentially a "cheat" for the MAC to network layer address
 * translation mechanism. It takes p_Interface and looks for the p_NodeCapID
 * node on that interface, returning tht interface ID by which p_NodeCapID
 * node can be reached. If the node is unreachable on this interface, this 
 * proc returns NWADDR_INVALID */
long nwNodeToInterfaceID(long p_NodeCapID, NWINTERFACE *p_Interface) {
  NWMEDIUM *l_Medium;
  NWINTERFACE *l_Int;
  if (!p_Interface) 
    return NWADDR_INVALID;

  if (p_NodeCapID==NWADDR_BROADCAST)
    return NWADDR_BROADCAST;
  if (p_NodeCapID==NWADDR_INVALID)
    return NWADDR_INVALID;

  l_Medium=p_Interface->iMedium;
  for(l_Int=l_Medium->mInterface;l_Int;l_Int=l_Int->iMediumNext) {
    if (l_Int->iNode->nNodeCap->nID==p_NodeCapID)
      return l_Int->iID;
  }
  return NWADDR_INVALID;
}

/* this proc is the reverse "cheat" for the nwNodeToInterface cheat */
long nwInterfaceToNodeID(long p_IntID, NWINTERFACE *p_Interface) {
  NWMEDIUM *l_Medium;
  NWINTERFACE *l_Int;
  if (!p_Interface) 
    return NWADDR_INVALID;

  if (p_IntID==NWADDR_BROADCAST)
    return NWADDR_BROADCAST;
  if (p_IntID==NWADDR_INVALID)
    return NWADDR_INVALID;

  l_Medium=p_Interface->iMedium;
  for(l_Int=l_Medium->mInterface;l_Int;l_Int=l_Int->iMediumNext) {
    if (l_Int->iID==p_IntID)
      return l_Int->iNode->nNodeCap->nID;
  }
  return NWADDR_INVALID;
}

void nwMediumEventStateOK(void *p_Parameter) {
  if (p_Parameter) {
    ((NWMEDIUM*)(p_Parameter))->mState=NWSTATE_OK;
    namLinkUp((NWMEDIUM*)p_Parameter);
  }
}
void nwMediumEventStateFail(void *p_Parameter) {
  if (p_Parameter) {
    ((NWMEDIUM*)(p_Parameter))->mState=NWSTATE_FAILURE;
    namLinkDown((NWMEDIUM*)p_Parameter);
  }
}

void nwSubNetConvergenceReportOn(void *p_Parameter) {
  if (p_Parameter) {
    ((NWSUBNETWORK*)(p_Parameter))->sReportConvergence=1;
  }
}
void nwSubNetConvergenceReportOff(void *p_Parameter) {
  if (p_Parameter) {
    ((NWSUBNETWORK*)(p_Parameter))->sReportConvergence=0;
  }
}


