#ifndef _NETWORK_H_
#define _NETWORK_H_
/* network.h
 * This file includes the structures for the network simulation */

/*** Definitions & Constants ***/
#define NWMETER 1000000  /* # virtual positions per meter */
#define NW_M(x) (x*NWMETER)
#define NW_MM(x) (x*(NWMETER/1000))

#define NWTICK      10000  /* # ticks per second */
#define NW_S(x)  (x*NWTICK)
#define NW_MS(x) (x*(NWTICK/1000))

/* Special Addresses (ID) (applies to nodes & interfaces ) */
#define NWADDR_BROADCAST -1
#define NWADDR_INVALID   -2

/* nDeliverError Error Codes */
#define NWDELERR_NOACK    1  /* did not receive ACK from receiver over medium */

#define NWNAMELEN        100 /* max length of a name */

/* used by NWSTATE */
#define NWSTATE_INVALID   0
#define NWSTATE_OK        1
#define NWSTATE_FAILURE   2

/* for use with xxxStatPrint() functions */
#define NWSPFLAG_PRINT (1<<0)
#define NWSPFLAG_CLEAR (1<<1)

struct NWNETWORK;
struct NWSUBNETWORK;
struct NWPACKET;
struct NWNODECAP;
struct NWNODE;
struct NWINTERFACE;
struct NWMEDIUM;

typedef struct NWNODECAP {
  long nID;                            /* ID */
  long nX,nY;                          /* coordinates */
  long nParam[10];                     /* Extra (undefined) parameters for nodecap operation */
  void (*nInitNodeCap)(struct NWNODECAP *p_NodeCap);
  void (*nTickNodeCap)(struct NWNODECAP *p_NodeCap); /* function to call each tick time */
  void (*nKillNodeCap)(struct NWNODECAP *p_NodeCap);
  struct NWPACKET *nPacketBuffer; /* pending packet to be sent.    */
  struct NWNODECAP *nPrev,*nNext;      /* linked list */
  struct NWNETWORK *nNet;              /* parent network */
  void *nGeneral;                      /* general purpose memory for use
                                          by the nTickNodeCap function   */
} NWNODECAP;

typedef struct NWNODE {
  struct NWNODECAP *nNodeCap;          /* corresponding node cap */
  struct NWSUBNETWORK *nSubNet;        /* parent sub network */
  struct NWNODE *nPrev,*nNext;         /* linked list */
  struct NWINTERFACE *nInterface;      /* linked-list of attached interfaces */
  void (*nInitNode)(struct NWNODE *p_Node);
  void (*nTickNode)(struct NWNODE *p_Node);
  void (*nKillNode)(struct NWNODE *p_Node);
  void (*nRcvPacket)(struct NWINTERFACE *p_Interface, struct NWPACKET *p_Packet);
  void (*nDeliveryError)(struct NWINTERFACE *p_Interface,struct NWPACKET *p_Packet,long p_ErrorCode);
  void *nGeneral;
} NWNODE;

typedef struct NWINTERFACE {
  long iID;
  struct NWNODE *iNode;                /* node this interface belongs to */
  struct NWINTERFACE *iNodePrev,*iNodeNext; /* linked list of other interfaces on this node */
  struct NWMEDIUM *iMedium;    /* interface handler (wireless..) */
  struct NWINTERFACE *iMediumPrev,*iMediumNext; /* linked list of other interfaces on this medium */
  void *iGeneral;
} NWINTERFACE;

typedef struct NWMEDIUM {
  long mID;
  long mState;
  char mName[NWNAMELEN];               /* medium name             */
  struct NWSUBNETWORK *mSubNetwork;
  struct NWINTERFACE *mInterface;      /* linked list of interfaces connected to this medium */
  struct NWMEDIUM *mPrev,*mNext;   /* linked list of technologies */
  long mBaudMultiplier;            /* mBaudIncrementer bytes are transmitted each mBaudMultiplier ticks */
  long mBaudIncrementer;           /* # bytes transmitted each mBaudMultiplier ticks. */
  float mBaud;                 /* baud rate in bits per second */
  float mPropagationDelay;     /* Propagation Delay in seconds */
  void (*mInitMedium)(struct NWMEDIUM *p_Medium); /* Init network structs  */
  void (*mTickMedium)(struct NWMEDIUM *p_Medium); /* do network operations  */
  void (*mKillMedium)(struct NWMEDIUM *p_Medium); /* do network operations  */
  void (*mKillInterface)(struct NWINTERFACE *p_Interface); 
  unsigned long (*mSendPacket)(struct NWINTERFACE *p_Interface, struct NWPACKET *p_Packet,long  p_Destination); 
  long (*mMinHop)(struct NWINTERFACE *p_SrcInt,long p_DstIntID);
  void *mGeneral;
} NWMEDIUM;

/* a topology is a connection scheme. So it could be "give each node
 * a radio". Or it could be "connect each node to two of its neighbors", 
 * or even a combination of two or more network technologies */
typedef struct NWSUBNETWORK {
  char sName[NWNAMELEN];               /* name of subnetwork   */
  int sFile;                           /* output .nam file     */
  struct NWNETWORK *sNet;              /* parent network       */
  struct NWNODE    *sNode;             /* linked list of nodes */
  struct NWMEDIUM  *sMedium;           /* list of tech's used in net   */
  struct NWSUBNETWORK *sPrev,*sNext;   /* linked list          */
  int sReportConvergence;              /* non-zero if convergence reporting is requested */
} NWSUBNETWORK;

/* each network has a parent NWNETWORK structure. When nodes etc.   */
/* are allocated, they are allocated to a NWNETWORK. Things like    */
/* protocols, network topology, and packet generation schemes are   */
/* shared and "common" for any given network.                       */
typedef struct NWNETWORK {
  unsigned long nTick;
  NWNODECAP *nNodeCap;                 /* linked list of nodes         */
  NWSUBNETWORK *nSubNetwork;           /* linked list of net scenarios */ 
  unsigned long nPauseTime;            /* pause time for nodes */
  unsigned long nStopTime;             /* time to stop simulation */
} NWNETWORK;

typedef struct NWPACKET {
  long            pID;           /* unique packet ID */
  unsigned long   pCreateTick;   /* tick when packet was created */
  long            pSourceID;     /* source nodecap */
  long            pDestID;       /* destination nodecap */
  unsigned long   pHops;         /* # hops in route */
  struct NWPACKET *pPrev,*pNext; /* if net/phys layers want to use it. */
  long            pTimeout;      /* if the net/phys layer want to use it. */
  long            pHoldoff;      /* if the net/phys layer want to use it. */
  long            pNamConversationID; /* set by the PROTOCOL - used by NAM */
  long            pNamColour;    /* set by the PROTOCOL - used by NAM */
  char            pNamString[20]; /* set by the PROTOCOL - used by NAM */
  /* NOTE: while the packet is in progress, the protocols are only 
   * allowed to look at the following information:                 */
  /* (other than to increment pHops)                               */
  long            pSourceIntID;  /* filled by sending *INTERFACE* */
  long            pDestIntID;    /* filled by sending *INTERFACE*, specified by sending *PROTOCOL* */
  unsigned long   pDataLen;      /* filled by sending *PROTOCOL* */
  unsigned char  *pData;         /* packet data - *PROTOCOL* */
  unsigned long   pPayloadLen;   /* filled by *APPLICATION* layer - we don't actually specify data, just how much there is. */
} NWPACKET;

NWNODECAP *nwAllocNodeCap(NWNETWORK *p_Net);
void nwFreeNodeCap(NWNODECAP *p_NodeCap);

NWNODE *nwAllocNode (NWSUBNETWORK *p_SubNet);
void nwFreeNode(NWNODE *p_Node);

NWINTERFACE *nwAllocInterface(NWNODE *p_Node,NWMEDIUM *p_Medium);
void nwFreeInterface(NWINTERFACE *p_Interface);

NWMEDIUM *nwAllocMedium(NWSUBNETWORK *p_SubNet);
void nwFreeMedium(NWMEDIUM *p_Medium);

NWSUBNETWORK *nwAllocSubNetwork(NWNETWORK *p_Net);
void nwFreeSubNetwork(NWSUBNETWORK *p_SubNet);

NWNETWORK *nwAllocNetwork(void);
void nwFreeNetwork(NWNETWORK *p_Net);

NWPACKET *nwAllocPacket(void);
void nwFreePacket(NWPACKET *p_Packet);

int nwInitNetwork(NWNETWORK *p_Net);
void nwTickNetwork(NWNETWORK *p_Net);
NWPACKET *nwSourcePacket(NWNODECAP *p_NodeCap, unsigned long p_PayloadLen, long p_DestinationID);
NWPACKET *nwCopyPacket(NWPACKET *p_Packet,int p_CopyData);
long nwNodeToInterfaceID(long p_NodeCapID, NWINTERFACE *p_Interface);
long nwInterfaceToNodeID(long p_IntID, NWINTERFACE *p_Interface);

/* Standard Event handlers */
void nwMediumEventStateOK(void *p_Parameter);
void nwMediumEventStateFail(void *p_Parameter);
void nwSubNetConvergenceReportOn(void *p_Parameter);
void nwSubNetConvergenceReportOff(void *p_Parameter);

/* Other utilities */
void nwBaudToMultIncr(NWMEDIUM *p_Medium);
void nwMultIncrToBaud(NWMEDIUM *p_Medium);
float nwTicksPerByte(NWMEDIUM *p_Medium);

#endif /* _NETWORK_H_ */

