/* schedule.c
 * Identifier: sch
 *
 * This module implements a scheduler for events such as link/node 
 * status modification events. However, since it operates using
 * simple callbacks, it can be used to schedule just about anything
 * you care to make a callback function for. 
 */

#include<stdlib.h>
#include"memmgr.h"

typedef struct SCHEDULED_ITEM {
  long sExpire; /* tick number of when this sucker expires */
  long sRepeat; /* if non-zero, this event will auto-reschedule with this period */
  void (*sCallback)(void* p_Parameter); /* callback - will be passed the sParameter data */
  void *sParameter; /* data to be passed to the sCallback function */
  struct SCHEDULED_ITEM *sNext; /* stored as a linked list, ordered according to sExpire */
} SCHEDULED_ITEM;

SCHEDULED_ITEM *g_schSchedule=NULL; /* This is our master scheduler linked list placeholder */

void schInsertEvent(SCHEDULED_ITEM *p_Item) {
  SCHEDULED_ITEM *l_current,*l_previous;
  if (p_Item) {
    /* let's insert this sucker into our event schedule */
    l_previous=NULL;
    for (l_current=g_schSchedule;
      (l_current)&&(l_current->sExpire<p_Item->sExpire);
      l_previous=l_current,l_current=l_current->sNext);
    p_Item->sNext=l_current;
    if (l_previous) {
      l_previous->sNext=p_Item;
    } else {
      /* this event goes at the start */
      g_schSchedule=p_Item;
    }
  }
}

void schScheduleTick(long p_Tick) {
  SCHEDULED_ITEM *l_t;
  /* check if we need to operate our scheduled item. Note we can do this quickly because
   * our events are sorted according to sExpire */
  while((g_schSchedule)&&(g_schSchedule->sExpire<=p_Tick)) {
    /* call event callback */
    (*(g_schSchedule->sCallback))(g_schSchedule->sParameter);
    /* now free/reinsert the scheduled item */
    l_t=g_schSchedule;
    g_schSchedule=g_schSchedule->sNext;
    if (l_t->sRepeat>0) {
      l_t->sExpire+=l_t->sRepeat;
      schInsertEvent(l_t);
    } else {
      mmuFree(l_t);
    }
  }
}


/* NOTE NOTE NOTE: the calling event is responsible for freeing sParameter if required */
void schScheduleEvent(long p_Expire,long p_Repeat,void (*p_Callback)(),void *p_Parameter) {
  SCHEDULED_ITEM *l_e;
  l_e=mmuAlloc(sizeof(SCHEDULED_ITEM));
  if (l_e) {
    l_e->sExpire=p_Expire;
    l_e->sRepeat=p_Repeat;
    l_e->sCallback=p_Callback;
    l_e->sParameter=p_Parameter;
    schInsertEvent(l_e);
  }
}


