/* wire.c
 * identifier: wm
 * 
 * This module implements a wired medium, much like a I2C or ethernet bus.
 */

/* Uncomment this for noisy/debug messages */
/* #define WMDEBUG */

#include<stdio.h>
#include<stdlib.h>
#include"memmgr.h"
#include"network.h"
#include"wire.h"
#include"nam.h"

/* Settings affecting wire.c behavior */
#define WM_RETRANS   1        /* # transmissions for failed (non-acked) packets eg: 3=1 orig trans + 3 retrans */

#define WM_BAUD             100000     /* bits per second */
#define WM_SENDHOLDOFF_MIN  NW_MS(2)  /* holdoff time after transmitting a pk */
#define WM_SENDHOLDOFF_MAX  NW_MS(6)
#define WM_RCVHOLDOFF_MIN   NW_MS(1)  /* Holdoff time after receiving a pk */
#define WM_RCVHOLDOFF_MAX   NW_MS(3)

#define WM_P_BASE         1000 /* Probability Base */
#define WM_P_DESTFAILURE   990 /* probability a packet sent to a specific node will succeed getting to that node */
#define WM_P_BCASTFAILURE  990 /* probability a broadcast packet (or a packet not addressed to the receiver) will succeed */

void wmWireInitMedium(NWMEDIUM *p_Medium) {
  NWINTERFACE *l_Interface;
  WMINTERFACEDATA *l_IntData;
  WMMEDIUMDATA *l_MedData;
  int l_i;

#ifdef WMDEBUG
  printf("Wire: wmWireInitMedium for medium %li\n",p_Medium->mID);
#endif
  /* first, go through each interface in our medium and alloc it a 
   * wire structure */
  for (l_Interface=p_Medium->mInterface;l_Interface; l_Interface=l_Interface->iMediumNext) {
    l_IntData=mmuAlloc(sizeof(WMINTERFACEDATA));
    l_Interface->iGeneral=(void*)l_IntData;
    l_IntData->iPacketBuffer=NULL;
    l_IntData->iNodePriority=0L;
    l_IntData->iNodeSendHoldoff=0L;
    l_IntData->iRetransmit=0L;
  }

  /* next, alloc a medium structure */
  l_MedData=mmuAlloc(sizeof(WMMEDIUMDATA));
  p_Medium->mGeneral=(void*)l_MedData;
  l_MedData->mSourceInt=NULL;
  l_MedData->mTicksToSend=0L;
  l_MedData->mTicksSent=0L;
  l_MedData->mSendHoldoff=0L;
  l_MedData->mCorrupted=0L;
  
  for (l_i=0;l_i<WM_S_COUNT;l_i++) {
    l_MedData->mStats[l_i]=0L;
  }
}

void wmWireTickMedium(NWMEDIUM *p_Medium) {
  NWINTERFACE *l_DstInt,*l_SrcInt;
  WMINTERFACEDATA *l_DstIntData,*l_SrcIntData;
  WMMEDIUMDATA *l_MedData;
  NWPACKET *l_Pk;
  long l_i;
  int l_sent;


  /* in this proc, we have to:
   *   0) Ensure the medium is up and drop any packets if it's gone down.
   *   1) Continue transmission of already in-flight packets 
   *   2) Select new senders from waiting senders            
   *   3) Deliver packets successfully sent 
   *
   * We do this using the following steps:
   *  STEP 1: If a packet is already being transmitted, increment mTicksSent.
   *          If mTicksSent>=mTicksToSend, we've finished transmitting the
   *          packet, so let's send it up to the protocol layer. 
   *          Exit.
   *  STEP 2: Medium is free - decrement holdoff values and, if appropriate, 
   *          start sending one of the pending new packets.
   */

  l_MedData=p_Medium->mGeneral;
  if (l_MedData->mSourceInt) {
    /* We are already sending a packet -- continue the process */
    l_MedData->mTicksSent++;
    l_SrcInt=l_MedData->mSourceInt;
    l_SrcIntData=l_SrcInt->iGeneral;
    l_Pk=l_SrcIntData->iPacketBuffer;

    /* for the next bit, l_sent is used to signal if we should drop our 
     * packet - l_sent=1=drop, l_sent=0=we're ok                        */
    l_sent=0;
    if ((p_Medium->mState!=NWSTATE_OK) && (!l_MedData->mCorrupted)) {
      l_sent=1;
      wmWireStatCount(p_Medium,WM_S_PACKETS_LOSTLINK);   /**** STATS ****/
    } else if (l_MedData->mTicksSent==(l_MedData->mTicksToSend/2)) {
      /* when we're half-way in flight, check for an error and drop
       * the packet if we've got one. */
      /* Check, using a somewhat random probability, if an error occurred. */
      /* I know for broadcast, you should have this probability of packet corruption, 
       * and a second probability (one per interface) that the packet was recieved, 
       * but hey, I'm not shooting for a gold medal simulator... I'm just trying to 
       * prove a point. */
      /* We do this calculation half-way through the packet's flight just 
       * so the animation in NAM is a little more explicit as to what's happening */
      if (l_Pk->pDestIntID==NWADDR_BROADCAST) {
        l_i=WM_P_BCASTFAILURE;
      } else {
        l_i=WM_P_DESTFAILURE;
      }
      if ((rand()%WM_P_BASE)>l_i) {
        /* oops - we've got an error! */
        l_sent=1;
        wmWireStatCount(p_Medium,WM_S_PACKETS_LOSTRANDOM); /**** STATS ****/
      }
    }
    if (l_sent) {
      /* we've just lost this packet. */
      l_MedData->mCorrupted=1;
      
      /* notify NAM */
      namPacketDiscard(p_Medium,l_Pk,l_MedData->mTicksToSend);
      if ((l_SrcIntData->iRetransmit>0)&&(l_Pk->pDestIntID!=NWADDR_BROADCAST)) {
        namPacketQueueAdd(l_SrcInt->iNode,l_Pk);
      }
    }

    if (l_MedData->mTicksSent>=l_MedData->mTicksToSend) {
      /* we're done sending this packet!!! */
#ifdef WMDEBUG
      printf("%8li Wire: Completed transmitting packet %li on medium %li from Interface %li to %li (src=%li dst=%li)\n",
        p_Medium->mSubNetwork->sNet->nTick,
        l_Pk->pID,
        p_Medium->mID,
        l_Pk->pSourceIntID,
        l_Pk->pDestIntID,
        l_Pk->pSourceID,
        l_Pk->pDestID);
#endif
      l_sent=0;
      if ((p_Medium->mState==NWSTATE_OK)&&(l_MedData->mCorrupted==0)) {
        /* phew, we didn't have an error! */
	/* Now we just have to search for interfaces connected to this medium that need to
	 * have the packet delivered */
        for (l_DstInt=p_Medium->mInterface;l_DstInt;l_DstInt=l_DstInt->iMediumNext) {
          if (l_DstInt!=l_SrcInt) {
            if ((l_DstInt->iID==l_Pk->pDestIntID)||(l_Pk->pDestIntID==NWADDR_BROADCAST)) {
              /* let's send this sucker */
	      (l_DstInt->iNode->nRcvPacket)(l_DstInt,l_Pk);
	      /* baboom - it's sent up to the protocol layer */
              /* ok, we've just delivered the packet to the interface. Now let's setup 
	       * a holdoff time during which the node can't transmit. */
	      l_DstIntData=l_DstInt->iGeneral;
              l_DstIntData->iNodeSendHoldoff=
                rand()%(WM_RCVHOLDOFF_MAX-WM_RCVHOLDOFF_MIN+1)+
                WM_RCVHOLDOFF_MIN;
              l_sent++;
	    }
	  }    
        }
      } 
#ifdef WMDEBUG
      else {
        printf("Wire:     Ouch! Packet got wacked in transmission. No reception occurred.\n");
      }
#endif
      /* at this point, if we sent our packet, l_sent>=1. Otherwise, transmission error */
      if (l_sent) {
        namPacketReceive(p_Medium,l_Pk,l_MedData->mTicksToSend);
        wmWireStatCount(p_Medium,WM_S_PACKETS_TRANSOK); /**** STATS ****/
      } else {
        /* we should rarely, if ever, end up here. An error would have to 
	 * occur on the very last tick, or the interface be addressed incorrectly or something */
        if (!l_MedData->mCorrupted)  {
          namPacketDiscard(p_Medium,l_Pk,l_MedData->mTicksToSend);
          if (l_SrcIntData->iRetransmit>0) {
            namPacketQueueAdd(l_SrcInt->iNode,l_Pk);
	  }
        }
      }
      if ((l_Pk->pDestIntID==NWADDR_BROADCAST)||(l_sent)) {
        /* we sent the packet, it was received (maybe), let's kill the packet & move on */
        nwFreePacket(l_Pk);
        l_SrcIntData->iPacketBuffer=NULL;
        l_SrcIntData->iNodePriority=0;
	/* setup a holdoff in the sending node so it can't transmit right away again */
        l_SrcIntData->iNodeSendHoldoff=
          rand()%(WM_SENDHOLDOFF_MAX-WM_SENDHOLDOFF_MIN+1)+
          WM_SENDHOLDOFF_MIN;
        l_SrcIntData->iRetransmit=0;
      } else {
        /* we're unicasting and our destination didn't get it! */
        /* Try to retransmit. */
        if (l_SrcIntData->iRetransmit>0) {
          wmWireStatCount(p_Medium,WM_S_RETRANSMISSIONS); /**** STATS ****/
          l_SrcIntData->iRetransmit--;
          /* retransmit -- ie: don't free the packet. */
	  /* We don't have to touch anything to cause a retransmission - 
	   * the interface will hold the packet (we just don't free it here) 
	   * and the interface will compete again for transmission rights
	   * for that packet again */
        } else {
          /* transmission error (no more retransmissions) - free the packet and 
	   * alert higher layer protocols. */
          (l_SrcInt->iNode->nDeliveryError)(l_SrcInt, l_Pk, NWDELERR_NOACK);
          l_SrcIntData->iPacketBuffer=NULL; 
          nwFreePacket(l_Pk);
        }
        l_SrcIntData->iNodePriority=0; /* have to re-contend for link if retransmitting --> no hogging */
	/* node has to delay a holdoff period */
        l_SrcIntData->iNodeSendHoldoff=
          rand()%(WM_SENDHOLDOFF_MAX-WM_SENDHOLDOFF_MIN+1)+
          WM_SENDHOLDOFF_MIN;
      } 
      /* delay sending of packets (medium is busy recovering) */
      l_MedData->mSendHoldoff=
        rand()%(WM_SENDHOLDOFF_MAX-WM_SENDHOLDOFF_MIN+1)+
        WM_SENDHOLDOFF_MIN;
      /* reset medium settings in preparation for the next transmission */
      l_MedData->mSourceInt=NULL;
      l_MedData->mTicksToSend=0;
      l_MedData->mTicksSent=0;
      l_MedData->mCorrupted=0L;
    }
  } else { /* we are not sending any packets... see if we should start... */
    if (l_MedData->mSendHoldoff) {
      l_MedData->mSendHoldoff--;
    } else {
      /* ok, let's pick a winner... who gets to send...*/
      /* We use l_DstInt just for the loop. l_SrcInt is our current "sending winner", and
       * l_i is our "sending winner's priority". */
      l_SrcInt=NULL;
      l_i=-1; 
      for (l_DstInt=p_Medium->mInterface;l_DstInt;l_DstInt=l_DstInt->iMediumNext) {
        l_DstIntData=l_DstInt->iGeneral;
	if ((l_DstIntData->iPacketBuffer)&&(!l_DstIntData->iNodeSendHoldoff)) {
	  if (l_DstIntData->iNodePriority>l_i) {
	    /* This is it! Our Current Winner! */
	    l_SrcInt=l_DstInt;
	    l_i=l_DstIntData->iNodePriority;
	  }
	}
      }
      if (l_SrcInt) {
        /* We've found a winner!!! Initiate Transmission. */
        /* Transmission will actually start NEXT tick; we just prepare it here. */
        wmWireStatCount(p_Medium,WM_S_PACKETS_SENT); /**** STATS ****/
	l_SrcIntData=l_SrcInt->iGeneral;
        l_MedData->mSourceInt=l_SrcInt;
	/* NOTE: the '4' at the start is for the MAC address - 32-bits */
	l_MedData->mTicksToSend=
	  (4 + l_SrcIntData->iPacketBuffer->pDataLen + l_SrcIntData->iPacketBuffer->pPayloadLen)
	  * nwTicksPerByte(p_Medium)
	  + (p_Medium->mPropagationDelay*NWTICK);
	l_MedData->mTicksSent=0;
	/* and we're ready to boogey */
#ifdef WMDEBUG
        l_Pk=l_SrcIntData->iPacketBuffer;
        printf("%8li Wire: Begin transmitting packet %li on medium %li from Interface %li/%li to %li (src node =%li dst node=%li)\n",
          p_Medium->mSubNetwork->sNet->nTick,
          l_Pk->pID,
          p_Medium->mID,
          l_Pk->pSourceIntID,l_SrcInt->iID,
          l_Pk->pDestIntID,
          l_Pk->pSourceID,
          l_Pk->pDestID);
#endif
        /* Send notification to the NAM file */
	namPacketQueueRemove(l_SrcInt->iNode,l_SrcIntData->iPacketBuffer);
	namPacketTransmit(p_Medium,l_SrcIntData->iPacketBuffer,l_MedData->mTicksToSend);
      }
    }
  }
  /* oh, and before we leave, increase the send priority of any pending packets */
  /* We can decrement any holdoff values at the same time */
  for (l_SrcInt=p_Medium->mInterface;l_SrcInt;l_SrcInt=l_SrcInt->iMediumNext) {
    l_SrcIntData=l_SrcInt->iGeneral;
    if ((l_SrcIntData->iPacketBuffer) &&(l_SrcInt!=l_MedData->mSourceInt))
      l_SrcIntData->iNodePriority++;
    if (l_SrcIntData->iNodeSendHoldoff)
      l_SrcIntData->iNodeSendHoldoff--;
  }
  /* that's it! */
}


void wmWireKillMedium(NWMEDIUM *p_Medium) {
  NWINTERFACE *l_Interface;
  /* first, go through each interface in our medium and free it's 
   * wire structure */
  for (l_Interface=p_Medium->mInterface;l_Interface;
    l_Interface=l_Interface->iMediumNext) {
    mmuFree(l_Interface->iGeneral);
    l_Interface->iGeneral=NULL;
  }

  /* next, free medium structure */
  mmuFree(p_Medium->mGeneral);
}

void wmWireKillInterface(NWINTERFACE *p_Interface) {
  mmuFree(p_Interface->iGeneral);
  p_Interface->iGeneral=NULL;
} 

unsigned long wmWireSendPacket(NWINTERFACE *p_Interface, NWPACKET *p_Packet, long p_DestinationID ) {
  WMINTERFACEDATA *l_IntData;
  WMMEDIUMDATA *l_MedData;

  l_IntData=p_Interface->iGeneral;
  l_MedData=p_Interface->iMedium->mGeneral;
  if (l_IntData->iPacketBuffer)
    return 0L;
  if ((p_Packet)&&(p_DestinationID!=NWADDR_INVALID)) {
    mmuSanityCheck();
    l_IntData->iPacketBuffer=nwCopyPacket(p_Packet,1);
    mmuSanityCheck();
    l_IntData->iPacketBuffer->pSourceIntID=p_Interface->iID;
    l_IntData->iPacketBuffer->pDestIntID=p_DestinationID;
    l_IntData->iRetransmit=WM_RETRANS;
    namPacketQueueAdd(p_Interface->iNode,p_Packet);
    wmWireStatCount(p_Interface->iMedium,WM_S_PACKETS_SOURCE); /**** STATS ****/
  }
  return 1L;
}


/* Increments a stats counter - p_Count should be one of the WM_S_* constants from wire.h */
void wmWireStatCount(NWMEDIUM *p_Medium,int p_Count) {
  WMMEDIUMDATA *l_MedData;
  l_MedData=p_Medium->mGeneral;
  l_MedData->mStats[p_Count]++;
}

/* Prints out the stats from the specified subnet */
void wmWireStatPrint(NWSUBNETWORK *p_SubNet,int p_Flag) {
  NWMEDIUM *l_Medium;
  WMMEDIUMDATA *l_MedData;
  long l_total[WM_S_COUNT];
  int l_i;

  for (l_i=0;l_i<WM_S_COUNT;l_i++)
    l_total[l_i]=0L;

  for (l_Medium=p_SubNet->sMedium;l_Medium;l_Medium=l_Medium->mNext) {
    l_MedData=l_Medium->mGeneral;
    for (l_i=0;l_i<WM_S_COUNT;l_i++) {
      l_total[l_i]+=l_MedData->mStats[l_i];
      if (p_Flag&NWSPFLAG_CLEAR)  {
        l_MedData->mStats[l_i]=0L;
      }
    }
  }

  if (p_Flag&NWSPFLAG_PRINT) {
    printf("                            #Packets  #Transms  LostRndm  LostLink  #TransOK  #Retrans\n");
    printf("% 3li.%04li STATS medium wire  ",p_SubNet->sNet->nTick/NWTICK,p_SubNet->sNet->nTick%NWTICK);
    for (l_i=0;l_i<WM_S_COUNT;l_i++)
      printf("%8li  ",l_total[l_i]);
    printf("\n");
  }
}


