/* wireless.c
 * identifier: wt
 * 
 * This module implements a wireless medium (physical layer) for the
 * NetSim network simulator. 
 * 
 * General Characteristics:
 * When a node wishes to transmit a packet, it must first contend for
 * control of the local medium. I say local because the affect is
 * local to a circular perimiter (radius specified in iGeneral). A
 * node can send if:
 *  1) no node is currently sending within 2*radius 
 *     (this eliminates the invisible terminal problem and simulates
 *     how some MAC layer protocols will work)
 *  2) if other nodes within range also want to send, then a global
 *     priority value (also in iGeneral) is compared. The node 
 *     with the highest priority (ie: has been waiting the longest) gets
 *     to send. 
 *
 * NOTE: A sender can be delayed by its holdoff value. A random holdoff value
 * is given to each node when it a) receives a packet or b) sends a packet.
 * The holdoff value ranges are specified by the WT_???HOLDOFF_MIN/MAX 
 * constants below.
 *
 * While a packet is in flight, other nodes around it will "hear" it. 
 * If a node is within range of 2 senders, it will receive neither 
 * packet successfully (although it'll still be tagged as "receiving" till
 * both senders have finished). 
 *
 * A packet sent to a given destination will require an ACK signal
 * (actually I don't explicitly ACK in the code -- rather I check if
 * the destination node received the packet, and return a NACK signal
 * to the protocol level). Before the network layer is notified of a
 * failure to send, the wireless interface tries to resend the packet 
 * WT_RETRANS times. IE: If WT_RETRANS is 2, the packet will be transmitted
 * twice, and then a NWDELERR_NOACK signal will be sent to the protocol 
 * layer.
 *
 * NOTE that broadcast destination addresses don't require ACKs and 
 * therefore will never cause a network delivery error. 
 *
 * Each packet takes (pDataLen+pPayloadLen)*mBaudMultiplier ticks to 
 * be sent. 
 *
 * Packet transmission failure is simulated by a simple probability 
 * calculation. There are 2 failure values: 
 *   WT_P_DESTFAILURE  is the probability a receiver will successfully receive
 *                     a packet addressed to it. 
 *   WT_P_BCASTFAILURE is the probability a receiver of a broadcast packet
 *                     OR a packet not addressed to it will successfully
 *                     receive the packet. 
 *
 * BIG NOTE BIG NOTE BIG NOTE:
 * The wireless medium uses INTERFACE IDs, *not* NodeCap IDs. Therefore, 
 * your network protocol layer must do some kind of interface (MAC) layer
 * to network layer address translation. 
 */

#include<stdlib.h>
#include"memmgr.h"
#include"network.h"
#include"wireless.h"

#define WTDEF_RADIUS (NWMETER*3) /* default transmission radius */
#define WT_RETRANS   3        /* # transmissions for failed (non-acked) packets eg: 3=1 orig trans + 2 retrans */

#define WT_SENDHOLDOFF_MIN  10  /* holdoff time after transmitting a pk */
#define WT_SENDHOLDOFF_MAX  19
#define WT_RCVHOLDOFF_MIN    3  /* Holdoff time after receiving a pk */
#define WT_RCVHOLDOFF_MAX    9

#define WT_P_BASE         1000 /* Probability Base */
#define WT_P_DESTFAILURE   950 /* probability a packet sent to a specific node will succeed getting to that node */
#define WT_P_BCASTFAILURE  950 /* probability a broadcast packet (or a packet not addressed to the receiver) will succeed */

void wtWirelessInitMedium(NWMEDIUM *p_Medium) {
  NWINTERFACE *l_Interface;
  WTINTERFACEDATA *l_IntData;
  WTMEDIUMDATA *l_MedData;

  /* first, go through each interface in our medium and alloc it a 
   * wireless structure */
  for (l_Interface=p_Medium->mInterface;l_Interface;
    l_Interface=l_Interface->iMediumNext) {
    l_IntData=mmuAlloc(sizeof(WTINTERFACEDATA));
    l_Interface->iGeneral=(void*)l_IntData;
    l_IntData->iTransmissionRadius=WTDEF_RADIUS;
    l_IntData->iPacketBuffer=NULL;
    l_IntData->iNodePriority=0L;
    l_IntData->iNodeStatus=WT_STATUS_IDLE;
    l_IntData->iNodeCountDown=0L;
    l_IntData->iNodeCountUp=0L;
    l_IntData->iNodeRcvFromID=NWADDR_INVALID;
    l_IntData->iNodeSendHoldoff=0L;
    l_IntData->iRetransmit=0L;
  }

  /* next, alloc a medium structure */
  l_MedData=mmuAlloc(sizeof(WTMEDIUMDATA));
  p_Medium->mGeneral=(void*)l_MedData;
  l_MedData->mSomeVal=0L; /* unused - just here as a place maker in case we use this structure some time in the future */
}

void wtWirelessTickMedium(NWMEDIUM *p_Medium) {
  WTINTERFACEDATA *l_IntData,*l_DestIntData,*l_NewSendData;
  NWINTERFACE *l_Int,*l_DestInt,*l_NewSend;
  NWPACKET *l_Pk;
  long l_i,x,y,r,ok;
  double l_fx,l_fy,l_fr;

  /* in this proc, we have to:
   *   1) Continue transmission of already in-flight packets 
   *      (and detect collisions based on node movement)
   *   2) Select new senders from waiting senders            
   *   3) Deliver packets successfully sent 
   *
   * We will do this in several steps:
   *   Step 1: Go through and clear all SENDING nodes which are now
   *           done. This is signified by a zero Countdown value.
   *           At the same time, remove stale RcvFromID tags.
   *           Also go through and tag all non-sending nodes as IDLE.
   *           iNodeRcvFromID and countup values are unchanged.
   *   Step 2: Go through and, for each sending node, re-tag nodes
   *           within range as RECEIVING.
   *           NOTE: If a node gets taged by one sender, and then
   *           again by another sender, this is considered a collision.
   *           (ie: the ranges are overlapping and the signals mix).
   *           Therefore, the node should be set to RECEIVING, but
   *           its count-up value is reset to 0. This will generate an
   *           error when the packet is done its transmission.
   *           NOTE: The value of iNodeRcvFromID *SHOULD* be set. Also, 
   *           IFF RcvFromID is set, reset countup to 0. If it's already
   *           correct (from a previous iteration), increment it. :)
   *   Step 3: Go through and, for each node which is marked as IDLE, 
   *           clear its Countup and RcvFromID field. This resets nodes which 
   *           leave from a transmission area while a packet is in flight. 
   *   Step 4: Check for new Senders. Go through the list and, if a node
   *           has a non-NULL iPacketBuffer, check that node against
   *           all currently sending nodes. If it's not going to cause
   *           a collision against any existing senders, allow it to 
   *           join the sender election pool. The election pool is just
   *           the set of nodes which have a packet to send and are 
   *           also elligible. The node with the highest iNodePriority
   *           wins and gets to send. Note after each assignment, this
   *           step repeats itself until there is no new sender assigned.
   *           This step also has to subsequently go through the list
   *           and mark any new receiving nodes appropriately.
   *   Step 5: Transmit the byte. This is done by decrementing all
   *           iCountdown values in senders and incrementing all
   *           iCountup values in receivers. If an iCountdown value
   *           reaches 0, check all nodes within range and if their
   *           iCountup equals the sender's iNodeCountVal, then we 
   *           assume the node heard the whole packet and the packet is
   *           delivered to it. Note: the iNodeRcvFromID must match in order
   *           for the packet to be delivered, otherwise it's a collision.
   *   Step 6: there is no step 6; we're done. 
   */

   /* Ok, here we go! */
  /*
   *   Step 1: Go through and clear all SENDING nodes which are now
   *           done. This is signified by a zero Countdown value.
   *           Also go through and tag all non-sending nodes as IDLE.
   *           iNodeRcvFromID and countup values are unchanged.
   */
  for (l_Int=p_Medium->mInterface;l_Int;l_Int=l_Int->iMediumNext) {
    l_IntData=l_Int->iGeneral;
    if (l_IntData->iNodeSendHoldoff) {
      l_IntData->iNodeSendHoldoff--;
    }
    if (l_IntData->iNodeStatus==WT_STATUS_SENDING) {
      if (l_IntData->iNodeCountDown==0) {
        l_IntData->iNodeStatus=WT_STATUS_IDLE;
        l_IntData->iNodeCountVal=0L;
        for (l_DestInt=p_Medium->mInterface;l_DestInt;l_DestInt=l_DestInt->iMediumNext) {
          l_DestIntData=l_DestInt->iGeneral;
          if (l_DestIntData->iNodeRcvFromID==l_Int->iID) {
            l_DestIntData->iNodeRcvFromID=NWADDR_INVALID;
            l_DestIntData->iNodeCountUp=0L;
            l_DestIntData->iNodeCountDown=0L;
            l_DestIntData->iNodeCountVal=0L;
          }
        }
      }
    } else {
      l_IntData->iNodeStatus=WT_STATUS_IDLE;
      if (l_IntData->iPacketBuffer) {
        l_IntData->iNodePriority++;
      } 
    }
  }

  /*   Step 2: Go through and, for each sending node, re-tag nodes
   *           within range as RECEIVING.
   *           NOTE: If a node gets taged by one sender, and then
   *           again by another sender, this is considered a collision.
   *           (ie: the ranges are overlapping and the signals mix).
   *           Therefore, the node should be set to RECEIVING, but
   *           its count-up value is reset to 0. This will generate an
   *           error when the packet is done its transmission.
   *           NOTE: The value of iNodeRcvFromID *SHOULD* be set. Also, 
   *           IFF RcvFromID is set, reset countup to 0. If it's already
   *           correct (from a previous iteration), increment it. :)
   */
  for (l_Int=p_Medium->mInterface;l_Int;l_Int=l_Int->iMediumNext) {
    l_IntData=l_Int->iGeneral;
    if (l_IntData->iNodeStatus==WT_STATUS_SENDING) {
      for (l_DestInt=p_Medium->mInterface;l_DestInt;l_DestInt=l_DestInt->iMediumNext) {
        if (l_DestInt!=l_Int) {
          l_DestIntData=l_DestInt->iGeneral;
          /* first, are we in range? */
          x=l_Int->iNode->nNodeCap->nX-l_DestInt->iNode->nNodeCap->nX;
          y=l_Int->iNode->nNodeCap->nY-l_DestInt->iNode->nNodeCap->nY;
          r=l_IntData->iTransmissionRadius;
          l_fx=x; l_fy=y; l_fr=r;
          /*if (((x*x)+(y*y))<=(r*r)) {*/
          if (((l_fx*l_fx)+(l_fy*l_fy))<=(l_fr*l_fr)) {
            if (l_DestIntData->iNodeStatus!=WT_STATUS_SENDING) {
              l_DestIntData->iNodeStatus=WT_STATUS_RECEIVING;
              if (l_DestIntData->iNodeRcvFromID==l_Int->iID) {
                l_DestIntData->iNodeCountUp++;
              } else {
                /* oh ho! We're fighting over this baby! */
                /* (collision) */
                l_DestIntData->iNodeRcvFromID=l_Int->iID;
                l_DestIntData->iNodeCountUp=0L;
              }
            } 
          } 
        }
      }
    }
  }

  /*   Step 3: Go through and, for each node which is marked as IDLE, 
   *           clear its Countup and RcvFromID field. This resets nodes which 
   *           leave from a transmission area while a packet is in flight. 
   */
  for (l_Int=p_Medium->mInterface;l_Int;l_Int=l_Int->iMediumNext) {
    l_IntData=l_Int->iGeneral;
    if (l_IntData->iNodeStatus==WT_STATUS_IDLE) {
      l_IntData->iNodeCountDown=0L;
      l_IntData->iNodeCountVal=0L;
      l_IntData->iNodeCountUp=0L;
      l_IntData->iNodeRcvFromID=NWADDR_INVALID;
    }
  }

  /*   Step 4: Check for new Senders. Go through the list and, if a node
   *           has a non-NULL iPacketBuffer, check that node against
   *           all currently sending nodes. If it's not going to cause
   *           a collision against any existing senders, allow it to 
   *           join the sender election pool. The election pool is just
   *           the set of nodes which have a packet to send and are 
   *           also elligible. The node with the highest iNodePriority
   *           wins and gets to send. Note after each assignment, this
   *           step repeats itself until there is no new sender assigned.
   *           This step also has to subsequently go through the list
   *           and mark any new receiving nodes appropriately.
   */

  do {
    /* cycle through checking each IDLE node */
    l_NewSend=NULL;
    for (l_Int=p_Medium->mInterface;l_Int;l_Int=l_Int->iMediumNext) {
      l_IntData=l_Int->iGeneral;
      if ((l_IntData->iNodeStatus==WT_STATUS_IDLE)&&
          (l_IntData->iPacketBuffer)&&(!l_IntData->iNodeSendHoldoff)) {
        /* First, check this sender's range from all other sender's ranges */
        ok=1;
        for (l_DestInt=p_Medium->mInterface;l_DestInt&&ok;l_DestInt=l_DestInt->iMediumNext) {
          if (l_DestInt!=l_Int) {
            l_DestIntData=l_DestInt->iGeneral;
            if (l_DestIntData->iNodeStatus==WT_STATUS_SENDING) {
              /* potential conflict -- check range */
              x=l_Int->iNode->nNodeCap->nX-l_DestInt->iNode->nNodeCap->nX;
              y=l_Int->iNode->nNodeCap->nY-l_DestInt->iNode->nNodeCap->nY;
              r=l_IntData->iTransmissionRadius;
              l_fx=x;l_fy=y;l_fr=r;
              /* note we check 2* radius */
              /*if (((x*x)+(y*y))<=(4*r*r)) {*/
              if (((l_fx*l_fx)+(l_fy*l_fy))<(4*l_fr*l_fr)) {
                ok=0;
              }
            }
          }
        }
        if (ok) {
          /* sender is clear! */
          /* compare priorities */
          if (l_NewSend) {
            if (((WTINTERFACEDATA*)(l_Int->iGeneral))->iNodePriority>
                ((WTINTERFACEDATA*)(l_NewSend->iGeneral))->iNodePriority) {
              l_NewSend=l_Int;
            } /* else node is lower priority - doesn't get the send */
          } else {
            l_NewSend=l_Int;
          }
        }
      }
    }
    if (l_NewSend) {
      /* we've got to make a new sender!!! */
      l_NewSendData=l_NewSend->iGeneral;
      l_NewSendData->iNodePriority=0L;
      l_NewSendData->iNodeStatus=WT_STATUS_SENDING;
      l_NewSendData->iNodeCountDown=((l_NewSendData->iPacketBuffer->pDataLen+
        l_NewSendData->iPacketBuffer->pPayloadLen) * 
        l_NewSend->iMedium->mBaudMultiplier) / 
        l_NewSend->iMedium->mBaudIncrementer + 1;
      l_NewSendData->iNodeCountVal=l_NewSendData->iNodeCountDown;
      if (!(l_NewSendData->iRetransmit))
        l_NewSendData->iRetransmit=WT_RETRANS;

      /*printf("Wireless: Packet %li being transmitted\n",
        l_NewSendData->iPacketBuffer->pID);*/

      /* now mark every node in range as receiver */
      for (l_DestInt=p_Medium->mInterface;l_DestInt;l_DestInt=l_DestInt->iMediumNext) {
        if (l_DestInt!=l_NewSend) {
          l_DestIntData=l_DestInt->iGeneral;
          if (l_DestIntData->iNodeStatus==WT_STATUS_IDLE) {
            x=l_NewSend->iNode->nNodeCap->nX-l_DestInt->iNode->nNodeCap->nX;
            y=l_NewSend->iNode->nNodeCap->nY-l_DestInt->iNode->nNodeCap->nY;
            r=l_NewSendData->iTransmissionRadius;
            /*if (((x*x)+(y*y))<=(r*r)) {*/
            l_fx=x; l_fy=y; l_fr=r;
            if (((l_fx*l_fx)+(l_fy*l_fy))<=(l_fr*l_fr)) {
              /* this node is now a receiver */
              l_DestIntData->iNodeStatus=WT_STATUS_RECEIVING;
              l_DestIntData->iNodeCountUp=1L;
              l_DestIntData->iNodeRcvFromID=l_NewSend->iID;
            } 
          } 
        }
      }
    }
  } while(l_NewSend);

  /*   Step 5: Transmit the byte. This is done by decrementing all
   *           iCountdown values in senders and incrementing all
   *           iCountup values in receivers. If an iCountdown value
   *           reaches 0, check all nodes within range and if their
   *           iCountup equals the sender's iNodeCountVal, then we 
   *           assume the node heard the whole packet and the packet is
   *           delivered to it. Note: the iNodeRcvFromID must match in order
   *           for the packet to be delivered, otherwise it's a collision.
   */
  for (l_Int=p_Medium->mInterface;l_Int;l_Int=l_Int->iMediumNext) {
    l_IntData=l_Int->iGeneral;
    if (l_IntData->iNodeStatus==WT_STATUS_SENDING) {
      l_IntData->iNodeCountDown--;
      if (!l_IntData->iNodeCountDown) {
        l_IntData->iNodeSendHoldoff=
          rand()%(WT_SENDHOLDOFF_MAX-WT_SENDHOLDOFF_MIN+1)+
          WT_SENDHOLDOFF_MIN;
        /* deliver packet to all nodes that received it! */
        l_Pk=l_IntData->iPacketBuffer;
        l_IntData->iPacketBuffer=NULL;
        ok=0;
        /* first, let's broadcast it to everyone in hearing distance */
        for (l_DestInt=p_Medium->mInterface;l_DestInt;l_DestInt=l_DestInt->iMediumNext) {
          if (l_DestInt!=l_Int) {
            l_DestIntData=l_DestInt->iGeneral;
            if (l_DestIntData->iNodeStatus==WT_STATUS_RECEIVING) {
              if (l_DestIntData->iNodeRcvFromID==l_Int->iID) {
                if (l_DestIntData->iNodeCountUp==l_IntData->iNodeCountVal) {
                  if (l_DestInt->iID==l_Pk->pDestIntID) {
                    l_i=WT_P_DESTFAILURE;
                  } else {
                    l_i=WT_P_BCASTFAILURE;
                  }
                  /* yep! let's send this sucker! */
                  /*printf("Wireless: Node %li receiving packet %li (p:from  %li/%li to %li/%li)\n",
                    l_DestInt->iNode->nNodeCap->nID,
                    l_Pk->pID,
                    l_Pk->pSourceID,
                    l_Pk->pSourceIntID,
                    l_Pk->pDestID,
                    l_Pk->pDestIntID);*/
                  if ((rand()%WT_P_BASE)<l_i) {
                    (l_DestInt->iNode->nRcvPacket)(l_DestInt,l_Pk);
                    l_DestIntData->iNodeSendHoldoff=
                      rand()%(WT_RCVHOLDOFF_MAX-WT_RCVHOLDOFF_MIN+1)+
                      WT_RCVHOLDOFF_MIN;
                    /* is this our real destination? */
                    if (l_DestInt->iID==l_Pk->pDestIntID) {
                      ok=1;
                    }
                  }  /* else packet transmission error */
                } 
              } 
            } 
          }
        }
        l_IntData->iNodeCountVal=0L;
        if ((l_Pk->pDestIntID==NWADDR_BROADCAST)||(ok)) {
          /* we sent the packet, it was received (maybe), let's kill the packet & move on */
          nwFreePacket(l_Pk);
          l_IntData->iRetransmit=0;
        } else {
          /* we're unicasting and our destination didn't get it! */
          /* Try to retransmit. */
          l_IntData->iRetransmit--;
          if (l_IntData->iRetransmit) {
            /* retransmit -- ie: don't free the packet. */
            l_IntData->iPacketBuffer=l_Pk;
          } else {
            /* transmission error */
            (l_Int->iNode->nDeliveryError)(l_Int, l_Pk, NWDELERR_NOACK);
            nwFreePacket(l_Pk);
          }
        }
      } 
    }
  }
}


void wtWirelessKillMedium(NWMEDIUM *p_Medium) {
  NWINTERFACE *l_Interface;
  /* first, go through each interface in our medium and free it's 
   * wireless structure */
  for (l_Interface=p_Medium->mInterface;l_Interface;
    l_Interface=l_Interface->iMediumNext) {
    mmuFree(l_Interface->iGeneral);
  }

  /* next, free medium structure */
  mmuFree(p_Medium->mGeneral);
}

void wtWirelessKillInterface(NWINTERFACE *p_Interface) {
  mmuFree(p_Interface->iGeneral);
} 

unsigned long wtWirelessSendPacket(NWINTERFACE *p_Interface, NWPACKET *p_Packet, long p_DestinationID ) {
  WTINTERFACEDATA *l_IntData;
  l_IntData=p_Interface->iGeneral;
  if (l_IntData->iPacketBuffer)
    return 0L;
  if ((p_Packet)&&(p_DestinationID!=NWADDR_INVALID)) {
    mmuSanityCheck();
    l_IntData->iPacketBuffer=nwCopyPacket(p_Packet,1);
    mmuSanityCheck();
    l_IntData->iPacketBuffer->pSourceIntID=p_Interface->iID;
    l_IntData->iPacketBuffer->pDestIntID=p_DestinationID;
  }
  return 1L;
}

long wtWirelessMinHop(NWINTERFACE *p_SrcInt,long p_DstIntID) {
  NWINTERFACE *l_Int,*l_rInt;
  long l_Hops,l_Changed;
  double l_fx,l_fy,l_fr;
  
  if (p_SrcInt->iID==p_DstIntID)
    return 0;

  /* first, let's reset all interfaces on this medium */
  for (l_Int=p_SrcInt->iMedium->mInterface;l_Int;l_Int=l_Int->iMediumNext) {
    ((WTINTERFACEDATA*)(l_Int->iGeneral))->iTag=0;
  }

  /* next, let's seed the network by tagging our source */
  ((WTINTERFACEDATA*)(p_SrcInt->iGeneral))->iTag=2;

  /* finally, let's propagate our tag info till we reach our destination */
  l_Changed=1;
  l_Hops=0;
  while(l_Changed) {
    l_Changed=0;
    l_Hops++;
    for (l_Int=p_SrcInt->iMedium->mInterface;l_Int;l_Int=l_Int->iMediumNext) {
      /* Is this a new sender (ie: is it's tag == 2) */
      if ( ((WTINTERFACEDATA*)(l_Int->iGeneral))->iTag==2) {
        ((WTINTERFACEDATA*)(l_Int->iGeneral))->iTag=3; /* no longer a new sender */
        /* now we have to check against all our other INTs to see if we're in range */
        for (l_rInt=p_SrcInt->iMedium->mInterface;l_rInt;l_rInt=l_rInt->iMediumNext) {
          /* is it unreached so far? */
          if ( ((WTINTERFACEDATA*)(l_rInt->iGeneral))->iTag==0) {
            /* is it within range of our sender? */
            l_fx=l_Int->iNode->nNodeCap->nX-l_rInt->iNode->nNodeCap->nX;
            l_fy=l_Int->iNode->nNodeCap->nY-l_rInt->iNode->nNodeCap->nY;
            l_fr=((WTINTERFACEDATA*)(l_Int->iGeneral))->iTransmissionRadius;
            if (((l_fx*l_fx)+(l_fy*l_fy))<=(l_fr*l_fr)) {
              l_Changed=1;
              ((WTINTERFACEDATA*)(l_rInt->iGeneral))->iTag=1;
              if (l_rInt->iID==p_DstIntID) /* is this our destination? */
                return l_Hops;
            }
          }
        }
      }
    }
    /* now advance all our receivers (tag=1) into senders (tag=2) */
    for (l_Int=p_SrcInt->iMedium->mInterface;l_Int;l_Int=l_Int->iMediumNext) {
      /* Is this receiver (ie: is it's tag == 1) */
      if ( ((WTINTERFACEDATA*)(l_Int->iGeneral))->iTag==1)
        ((WTINTERFACEDATA*)(l_Int->iGeneral))->iTag=2;
    }
  }

  return 99999; /* return max value -- destination unreachable */
}

